"""Bases classes and functions for rendering maps or layout elements"""

import dataclasses
import copy
import abc
import typing
import collections.abc
import os
import pathlib

import momapy.drawing
import momapy.styling
import momapy.positioning
import momapy.geometry
import momapy.builder
import momapy.core
import momapy.core.builders
import momapy.core.elements
import momapy.core.map


def register_renderer(name, renderer_cls):
    """Register a renderer class.

    .. deprecated::
        Use `momapy.rendering.register_renderer()` instead.
    """
    import momapy.rendering

    momapy.rendering.register_renderer(name, renderer_cls)


def _detect_renderer(format_: str) -> str:
    import momapy.rendering

    for candidate in ["svg-native", "skia", "cairo"]:
        try:
            renderer_cls = momapy.rendering.get_renderer(candidate)
            if format_ in renderer_cls.supported_formats:
                return candidate
        except (ValueError, ImportError, ModuleNotFoundError):
            continue
    raise ValueError(f"No renderer available for format '{format_}'")


def render_layout_element(
    layout_element: momapy.core.elements.LayoutElement,
    file_path: str | os.PathLike,
    format_: str | None = None,
    renderer: str | None = None,
    style_sheet: momapy.styling.StyleSheet | None = None,
    to_top_left: bool = False,
):
    """Render a layout element to a file in the given format with the given registered renderer

    Args:
        layout_element: The layout element to render
        file_path: The output file path
        format_: The output format. If None, inferred from file extension.
        renderer: The registered renderer to use. If None, auto-detected based on format.
        style_sheet: An optional style sheet to apply before rendering
        to_top_left: Whether to move the layout element to the top left or not before rendering
    """
    render_layout_elements(
        layout_elements=[layout_element],
        file_path=file_path,
        format_=format_,
        renderer=renderer,
        style_sheet=style_sheet,
        to_top_left=to_top_left,
    )


def render_layout_elements(
    layout_elements: collections.abc.Collection[momapy.core.elements.LayoutElement],
    file_path: str | os.PathLike,
    format_: str | None = None,
    renderer: str | None = None,
    style_sheet: momapy.styling.StyleSheet | None = None,
    to_top_left: bool = False,
    multi_pages: bool = True,
):
    """Render a collection of layout elements to a file in the given format with the given registered renderer.

    Args:
        layout_elements: The layout elements to render
        file_path: The output file path
        format_: The output format. If None, inferred from file extension.
        renderer: The registered renderer to use. If None, auto-detected based on format.
        style_sheet: An optional style sheet to apply before rendering
        to_top_left: Whether to move the layout elements to the top left before rendering
        multi_pages: Whether to render each layout element on a separate page
    """
    if format_ is None:
        file_path_obj = pathlib.Path(file_path)
        format_ = file_path_obj.suffix[1:]
        if not format_:
            raise ValueError(
                "Cannot determine format from file path. Please specify format_ parameter."
            )
    if renderer is None:
        renderer = _detect_renderer(format_)

    def _prepare_layout_elements(layout_elements, style_sheet=None, to_top_left=False):
        bboxes = [layout_element.bbox() for layout_element in layout_elements]
        bbox = momapy.positioning.fit(bboxes)
        max_x = bbox.x + bbox.width / 2
        max_y = bbox.y + bbox.height / 2
        if style_sheet is not None or to_top_left:
            new_layout_elements = []
            for layout_element in layout_elements:
                if isinstance(layout_element, momapy.core.elements.LayoutElement):
                    new_layout_elements.append(
                        momapy.builder.builder_from_object(layout_element)
                    )
                elif isinstance(layout_element, momapy.core.builders.LayoutElementBuilder):
                    new_layout_elements.append(copy.deepcopy(layout_element))
            layout_elements = new_layout_elements
        if style_sheet is not None:
            if (
                not isinstance(style_sheet, collections.abc.Collection)
                or isinstance(style_sheet, str)
                or isinstance(style_sheet, momapy.styling.StyleSheet)
            ):
                style_sheets = [style_sheet]
            else:
                style_sheets = style_sheet
            style_sheets = [
                (
                    momapy.styling.StyleSheet.from_file(style_sheet)
                    if not isinstance(style_sheet, momapy.styling.StyleSheet)
                    else style_sheet
                )
                for style_sheet in style_sheets
            ]
            style_sheet = momapy.styling.combine_style_sheets(style_sheets)
            for layout_element in layout_elements:
                momapy.styling.apply_style_sheet(layout_element, style_sheet)
        if to_top_left:
            min_x = bbox.x - bbox.width / 2
            min_y = bbox.y - bbox.height / 2
            max_x -= min_x
            max_y -= min_y
            translation = momapy.geometry.Translation(-min_x, -min_y)
            for layout_element in layout_elements:
                for attr_name in ["group_transform", "transform"]:
                    if hasattr(layout_element, attr_name):
                        if getattr(layout_element, attr_name) is None:
                            setattr(
                                layout_element, attr_name, []
                            )
                        getattr(layout_element, attr_name).append(translation)
                        break
        return layout_elements, max_x, max_y

    from momapy.rendering import get_renderer

    renderer_cls = get_renderer(renderer)
    if not multi_pages:
        prepared_layout_elements, max_x, max_y = _prepare_layout_elements(
            layout_elements, style_sheet, to_top_left
        )
        renderer_instance = renderer_cls.from_file(file_path, max_x, max_y, format_)
        renderer_instance.begin_session()
        for prepared_layout_element in prepared_layout_elements:
            renderer_instance.render_layout_element(prepared_layout_element)
        renderer_instance.end_session()
    else:
        if layout_elements:
            layout_element = layout_elements[0]
            prepared_layout_elements, max_x, max_y = _prepare_layout_elements(
                [layout_element], style_sheet, to_top_left
            )
            renderer_instance = renderer_cls.from_file(file_path, max_x, max_y, format_)
            renderer_instance.begin_session()
            renderer_instance.render_layout_element(prepared_layout_elements[0])
            for layout_element in layout_elements[1:]:
                prepared_layout_elements, max_x, max_y = _prepare_layout_elements(
                    [layout_element], style_sheet, to_top_left
                )
                renderer_instance.new_page(max_x, max_y)
                renderer_instance.render_layout_element(prepared_layout_elements[0])
        else:
            renderer_instance = renderer_cls.from_file(file_path, 0, 0, format_)
        renderer_instance.end_session()


def render_map(
    map_: momapy.core.map.Map,
    file_path: str | os.PathLike,
    format_: str | None = None,
    renderer: str | None = None,
    style_sheet: momapy.styling.StyleSheet | None = None,
    to_top_left: bool = False,
):
    """Render a map to a file in the given format with the given registered renderer.

    Args:
        map_: The map to render
        file_path: The output file path
        format_: The output format. If None, inferred from file extension.
        renderer: The registered renderer to use. If None, auto-detected based on format.
        style_sheet: An optional style sheet to apply before rendering
        to_top_left: Whether to move the map to the top left before rendering

    Example:
        >>> from momapy.io.core import read
        >>> from momapy.rendering.core import render_map
        >>> # Read a map from file
        >>> result = read("path/to/map.sbgn")
        >>> sbgn_map = result.obj
        >>> # Render the map to SVG
        >>> render_map(sbgn_map, "output.svg")
    """
    render_maps([map_], file_path, format_, renderer, style_sheet, to_top_left)


def render_maps(
    maps: collections.abc.Collection[momapy.core.map.Map],
    file_path: str | os.PathLike,
    format_: str | None = None,
    renderer: str | None = None,
    style_sheet: momapy.styling.StyleSheet | None = None,
    to_top_left: bool = False,
    multi_pages: bool = True,
):
    """Render a collection of maps to a file in the given format with the given registered renderer.

    Args:
        maps: The maps to render
        file_path: The output file path
        format_: The output format. If None, inferred from file extension.
        renderer: The registered renderer to use. If None, auto-detected based on format.
        style_sheet: An optional style sheet to apply before rendering
        to_top_left: Whether to move the maps to the top left before rendering
        multi_pages: Whether to render each map on a separate page

    Example:
        >>> from momapy.io.core import read
        >>> from momapy.rendering.core import render_maps
        >>> # Read multiple maps from files
        >>> result1 = read("path/to/map1.sbgn")
        >>> first_map = result1.obj
        >>> result2 = read("path/to/map2.sbgn")
        >>> second_map = result2.obj
        >>> # Render both maps to a multi-page PDF
        >>> render_maps([first_map, second_map], "output.pdf", multi_pages=True)
    """
    layout_elements = [map_.layout for map_ in maps]
    render_layout_elements(
        layout_elements=layout_elements,
        file_path=file_path,
        format_=format_,
        renderer=renderer,
        style_sheet=style_sheet,
        to_top_left=to_top_left,
        multi_pages=multi_pages,
    )


@dataclasses.dataclass
class Renderer(abc.ABC):
    """Base class for renderers"""

    initial_values: typing.ClassVar[dict] = {
        "font_family": momapy.drawing.DEFAULT_FONT_FAMILY,
        "font_weight": 16.0,
    }
    font_weight_value_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FontWeight.NORMAL: 400,
        momapy.drawing.FontWeight.BOLD: 700,
    }

    @abc.abstractmethod
    def begin_session(self):
        """Begin a session"""
        pass

    @abc.abstractmethod
    def end_session(self):
        """End a session"""
        pass

    @abc.abstractmethod
    def new_page(self, width, height):
        """Make a new page"""
        pass

    @abc.abstractmethod
    def render_map(self, map_: momapy.core.map.Map):
        """Render a map"""
        pass

    @abc.abstractmethod
    def render_layout_element(self, layout_element: momapy.core.elements.LayoutElement):
        """Render a layout element"""
        pass

    @abc.abstractmethod
    def render_drawing_element(self, drawing_element: momapy.drawing.DrawingElement):
        """Render a drawing element"""
        pass

    @classmethod
    def get_lighter_font_weight(
        cls, font_weight: momapy.drawing.FontWeight | float
    ) -> float:
        """Return the boldest font weight lighter than the given font weight"""
        if isinstance(font_weight, momapy.drawing.FontWeight):
            font_weight = cls.font_weight_value_mapping.get(font_weight)
            if font_weight is None:
                raise ValueError(
                    f"font weight must be a float, {momapy.drawing.FontWeight.NORMAL}, or {momapy.drawing.FontWeight.BOLD}"
                )
        if font_weight > 700:
            new_font_weight = 700
        elif font_weight > 500:
            new_font_weight = 400
        else:
            new_font_weight = 100
        return new_font_weight

    @classmethod
    def get_bolder_font_weight(
        cls, font_weight: momapy.drawing.FontWeight | float
    ) -> float:
        """Return the lightest font weight bolder than the given font weight"""
        if isinstance(font_weight, momapy.drawing.FontWeight):
            font_weight = cls.font_weight_value_mapping.get(font_weight)
            if font_weight is None:
                raise ValueError(
                    f"font weight must be a float, {momapy.drawing.FontWeight.NORMAL}, or {momapy.drawing.FontWeight.BOLD}"
                )
        if font_weight < 400:
            new_font_weight = 400
        elif font_weight < 600:
            new_font_weight = 700
        else:
            new_font_weight = 900
        return new_font_weight


@dataclasses.dataclass
class StatefulRenderer(Renderer):
    """Base class for stateful renderers"""

    _current_state: dict = dataclasses.field(default_factory=dict)
    _states: list[dict] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        self._initialize_current_state()

    @abc.abstractmethod
    def self_save(self):
        """Save the internal state of the renderer.

        This method must be implemented by subclasses to save any internal
        state that is not part of the current state dictionary.
        """
        pass

    @abc.abstractmethod
    def self_restore(self):
        """Restore the internal state of the renderer.

        This method must be implemented by subclasses to restore any internal
        state that is not part of the current state dictionary.
        """
        pass

    @classmethod
    def _make_initial_current_state(cls):
        state = {}
        for (
            attr_name,
            attr_d,
        ) in momapy.drawing.PRESENTATION_ATTRIBUTES.items():
            if attr_name in cls.initial_values:
                attr_value = cls.initial_values[attr_name]
            else:
                attr_value = attr_d["initial"]
            state[attr_name] = attr_value
        return state

    def _initialize_current_state(self):
        state = self._make_initial_current_state()
        self.set_current_state(state)

    def save(self):
        """Save the current state"""
        self._states.append(copy.deepcopy(self.get_current_state()))
        self.self_save()

    def restore(self):
        """Set the current state to the last saved state"""
        if len(self._states) > 0:
            state = self._states.pop()
            self.set_current_state(state)
            self.self_restore()
        else:
            raise Exception("no state to be restored")

    def get_initial_value(self, attr_name: str) -> typing.Any:
        """Return the initial value for an attribute"""
        attr_value = self.initial_values.get(attr_name)
        if attr_value is None:
            attr_d = momapy.drawing.PRESENTATION_ATTRIBUTES[attr_name]
            attr_value = attr_d["initial"]
            if attr_value is None:
                attr_value = momapy.drawing.INITIAL_VALUES[attr_name]
        return attr_value

    def get_current_value(self, attr_name: str) -> typing.Any:
        """Return the current value for an attribute"""
        return self.get_current_state()[attr_name]

    def get_current_state(self) -> dict[str, typing.Any]:
        """Return the current state"""
        return self._current_state

    def set_current_value(self, attr_name: str, attr_value: typing.Any):
        """Set the current value for an attribute"""
        if attr_value is None:
            attr_d = momapy.drawing.PRESENTATION_ATTRIBUTES[attr_name]
            if not attr_d["inherited"]:
                attr_value = self.initial_values.get(attr_name)
                if attr_value is None:
                    attr_value = attr_d["initial"]
                if attr_value is None:
                    attr_value = momapy.drawing.INITIAL_VALUES[attr_name]
        if attr_name == "font_weight":
            if isinstance(attr_value, momapy.drawing.FontWeight):
                if (
                    attr_value == momapy.drawing.FontWeight.NORMAL
                    or attr_value == momapy.drawing.FontWeight.BOLD
                ):
                    attr_value = self.font_weight_value_mapping[attr_value]
                elif attr_value == momapy.drawing.FontWeight.BOLDER:
                    attr_value = self.get_bolder_font_weight(
                        self.get_current_value("font_weight")
                    )
                elif attr_value == momapy.drawing.FontWeight.LIGHTER:
                    attr_value = self.get_lighter_font_weight(
                        self.get_current_value("font_weight")
                    )
        if attr_value is not None:
            self._current_state[attr_name] = attr_value

    def set_current_state(self, state: dict):
        """Set the current state to the given state"""
        for attr_name, attr_value in state.items():
            self.set_current_value(attr_name, attr_value)

    def _get_state_from_drawing_element(self, drawing_element):
        """Return the presentation state from a drawing element.

        Args:
            drawing_element: The drawing element to extract state from

        Returns:
            A dictionary mapping attribute names to their values
        """
        state = {}
        for attr_name in momapy.drawing.PRESENTATION_ATTRIBUTES:
            state[attr_name] = getattr(drawing_element, attr_name)
        return state

    def set_current_state_from_drawing_element(
        self, drawing_element: momapy.drawing.DrawingElement
    ):
        """Set the current state to a state given by a drawing element"""
        state = self._get_state_from_drawing_element(drawing_element)
        self.set_current_state(state)



"""Class for rendering with Skia"""

import dataclasses
import typing
import typing_extensions
import math
import os

try:
    import skia
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        f"The '{__name__}' module requires the 'skia-python' package. "
        f"Install it with: pip install skia-python  "
        f"Or install momapy with skia support: pip install momapy[skia]"
    ) from e

import momapy.drawing
import momapy.geometry
import momapy.rendering.core


@dataclasses.dataclass(kw_only=True)
class SkiaRenderer(momapy.rendering.core.StatefulRenderer):
    """Renderer implementation using the Skia graphics library.

    This renderer supports multiple output formats including PDF, SVG, PNG,
    JPEG, and WebP. It provides hardware-accelerated rendering capabilities
    and advanced features like filters and effects.

    Attributes:
        canvas: The Skia canvas used for rendering

    Example:
        >>> from momapy.meta.nodes import Rectangle
        >>> import momapy.geometry
        >>> # Create a layout element to render
        >>> node = Rectangle(
        ...     position=momapy.geometry.Point(100, 100),
        ...     width=200,
        ...     height=100
        ... )
        >>> # Create renderer and render the element
        >>> renderer = SkiaRenderer.from_file("output.pdf", 800, 600, "pdf")
        >>> renderer.begin_session()
        >>> renderer.render_layout_element(node)
        >>> renderer.end_session()
    """

    supported_formats: typing.ClassVar[list[str]] = [
        "pdf",
        "svg",
        "png",
        "jpeg",
        "webp",
    ]
    _de_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.Group: "_render_group",
        momapy.drawing.Path: "_render_path",
        momapy.drawing.Text: "_render_text",
        momapy.drawing.Ellipse: "_render_ellipse",
        momapy.drawing.Rectangle: "_render_rectangle",
    }
    _pa_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.MoveTo: "_add_move_to",
        momapy.drawing.LineTo: "_add_line_to",
        momapy.drawing.CurveTo: "_add_curve_to",
        momapy.drawing.QuadraticCurveTo: "_add_quadratic_curve_to",
        momapy.drawing.ClosePath: "_add_close_path",
        momapy.drawing.EllipticalArc: "_add_elliptical_arc",
    }
    _tr_class_func_mapping: typing.ClassVar[dict] = {
        momapy.geometry.Translation: "_add_translation",
        momapy.geometry.Rotation: "_add_rotation",
        momapy.geometry.Scaling: "_add_scaling",
        momapy.geometry.MatrixTransformation: "_add_matrix_transformation",
    }
    _fe_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.DropShadowEffect: "_make_drop_shadow_effect",
        momapy.drawing.CompositeEffect: "_make_composite_effect",
        momapy.drawing.GaussianBlurEffect: "_make_gaussian_blur_effect",
        momapy.drawing.OffsetEffect: "_make_offset_effect",
        momapy.drawing.FloodEffect: "_make_flood_effect",
    }
    _fe_composite_comp_op_blendmode_mapping: typing.ClassVar[dict] = {
        momapy.drawing.CompositionOperator.OVER: skia.BlendMode.kSrcOver,
        momapy.drawing.CompositionOperator.IN: skia.BlendMode.kSrcIn,
        momapy.drawing.CompositionOperator.OUT: skia.BlendMode.kSrcOut,
        momapy.drawing.CompositionOperator.ATOP: skia.BlendMode.kSrcATop,
        momapy.drawing.CompositionOperator.XOR: skia.BlendMode.kXor,
        momapy.drawing.CompositionOperator.LIGHTER: skia.BlendMode.kLighten,
        momapy.drawing.CompositionOperator.ARTIHMETIC: None,
    }
    _fe_gaussian_blur_edgemode_tilemode_mapping: typing.ClassVar[dict] = {
        momapy.drawing.EdgeMode.WRAP: skia.TileMode.kMirror,
        momapy.drawing.EdgeMode.DUPLICATE: skia.TileMode.kClamp,
        None: skia.TileMode.kDecal,
    }
    _te_font_style_slant_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FontStyle.NORMAL: skia.FontStyle.Slant.kUpright_Slant,
        momapy.drawing.FontStyle.ITALIC: skia.FontStyle.Slant.kItalic_Slant,
        momapy.drawing.FontStyle.OBLIQUE: skia.FontStyle.Slant.kOblique_Slant,
    }
    canvas: skia.Canvas = dataclasses.field(metadata={"description": "A skia canvas"})
    _config: dict = dataclasses.field(default_factory=dict)
    _skia_typefaces: dict = dataclasses.field(default_factory=dict)
    _skia_fonts: dict = dataclasses.field(default_factory=dict)

    @classmethod
    def from_file(
        cls,
        file_path: str | os.PathLike,
        width: float,
        height: float,
        format_: typing.Literal["pdf", "svg", "png", "jpeg", "webp"] = "pdf",
    ) -> typing_extensions.Self:
        """Create a SkiaRenderer instance from a file path.

        Args:
            file_path: The output file path
            width: The width of the canvas
            height: The height of the canvas
            format_: The output format (pdf, svg, png, jpeg, or webp)

        Returns:
            A new SkiaRenderer instance

        Raises:
            ValueError: If the format is not supported

        Example:
            >>> renderer = SkiaRenderer.from_file("output.pdf", 800, 600, "pdf")
        """
        if format_ not in cls.supported_formats:
            raise ValueError(f"Unsupported format: {format_}")
        config = {}
        canvas = None
        if format_ == "pdf":
            stream = skia.FILEWStream(file_path)
            document = skia.PDF.MakeDocument(stream)
            canvas = document.beginPage(width, height)
            config["stream"] = stream
            config["document"] = document
        elif format_ in ["png", "jpeg", "webp"]:
            surface = skia.Surface(width=int(width), height=int(height))
            canvas = surface.getCanvas()
            config["surface"] = surface
            config["file_path"] = file_path
        elif format_ == "svg":
            stream = skia.FILEWStream(file_path)
            canvas = skia.SVGCanvas.Make((width, height), stream)
            config["stream"] = stream
        config["file_path"] = file_path
        config["width"] = width
        config["height"] = height
        config["format"] = format_
        return cls(canvas=canvas, _config=config)

    def begin_session(self):
        """Begin a rendering session.

        This method initializes the rendering context. For SkiaRenderer,
        no explicit initialization is needed beyond the canvas setup.
        """
        pass

    def end_session(self):
        """End the rendering session and save the output.

        This method finalizes the rendering, flushes the canvas, and saves
        the output to the file. The specific actions depend on the output format:
        - PDF: Ends the page and closes the document
        - PNG/JPEG/WebP: Takes a snapshot and saves the image
        - SVG: Flushes the stream
        """
        self.canvas.flush()
        format_ = self._config.get("format")
        if format_ == "pdf":
            self._config["document"].endPage()
            self._config["document"].close()
        elif format_ == "png":
            image = self._config["surface"].makeImageSnapshot()
            image.save(self._config["file_path"], skia.kPNG)
        elif format_ == "jpeg":
            image = self._config["surface"].makeImageSnapshot()
            image.save(self._config["file_path"], skia.kJPEG)
        elif format_ == "webp":
            image = self._config["surface"].makeImageSnapshot()
            image.save(self._config["file_path"], skia.kWEBP)
        elif format_ == "svg":
            del self.canvas
            self._config["stream"].flush()

    def new_page(self, width, height):
        """Create a new page in the output document.

        Args:
            width: The width of the new page
            height: The height of the new page

        Note:
            Only PDF format supports multiple pages. Other formats will ignore this call.
        """
        format_ = self._config.get("format")
        if format_ == "pdf":
            self._config["document"].endPage()
            canvas = self._config["document"].beginPage(width, height)
            self.canvas = canvas

    def render_map(self, map_):
        """Render a map to the output.

        Args:
            map_: The map to render
        """
        self.render_layout_element(map_.layout)

    def render_layout_element(self, layout_element):
        """Render a layout element to the output.

        Args:
            layout_element: The layout element to render
        """
        drawing_elements = layout_element.drawing_elements()
        for drawing_element in drawing_elements:
            self.render_drawing_element(drawing_element)

    def render_drawing_element(self, drawing_element):
        """Render a drawing element to the output.

        Args:
            drawing_element: The drawing element to render

        This method handles filters, transformations, and delegates to
        the appropriate rendering method based on the drawing element type.
        """
        self.save()
        self.set_current_state_from_drawing_element(drawing_element)
        self._add_transform_from_drawing_element(drawing_element)
        class_ = type(drawing_element)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        de_func = getattr(self, self._de_class_func_mapping[class_])
        filter = self.get_current_value("filter")
        if filter is not momapy.drawing.NoneValue:
            bbox = drawing_element.bbox()
            saved_canvas = self.canvas
            recorder = skia.PictureRecorder()
            canvas = recorder.beginRecording(
                skia.Rect.MakeXYWH(
                    bbox.north_west().x,
                    bbox.north_west().y,
                    bbox.width,
                    bbox.height,
                )
            )
            self.canvas = canvas
            de_func(drawing_element)
            picture = recorder.finishRecordingAsPicture()
            skia_paint = self._make_filter_paint(
                filter, drawing_element.get_filter_region()
            )
            self.canvas = saved_canvas
            self.canvas.drawPicture(picture, paint=skia_paint)
        else:
            de_func(drawing_element)
        self.restore()

    def self_save(self):
        """Save the Skia canvas state.

        This method saves the current state of the Skia canvas, including
        transformations and clipping regions.
        """
        self.canvas.save()

    def self_restore(self):
        """Restore the Skia canvas state.

        This method restores the Skia canvas to the state saved by the
        most recent call to self_save().
        """
        self.canvas.restore()

    def _make_stroke_paint(self):
        if self.get_current_value("stroke_dasharray") is not momapy.drawing.NoneValue:
            skia_path_effect = skia.DashPathEffect.Make(
                list(self.get_current_value("stroke_dasharray")),
                self.get_current_value("stroke_dashoffset"),
            )
        else:
            skia_path_effect = None
        skia_paint = skia.Paint(
            AntiAlias=True,
            Color4f=skia.Color4f(
                self.get_current_value("stroke").to_rgba(rgba_range=(0.0, 1.0))
            ),
            StrokeWidth=self.get_current_value("stroke_width"),
            PathEffect=skia_path_effect,
            Style=skia.Paint.kStroke_Style,
        )
        return skia_paint

    def _make_fill_paint(self):
        skia_paint = skia.Paint(
            AntiAlias=True,
            Color4f=skia.Color4f(
                self.get_current_value("fill").to_rgba(rgba_range=(0.0, 1.0))
            ),
            Style=skia.Paint.kFill_Style,
        )
        return skia_paint

    def _make_filter_paint(self, filter_, filter_region):
        dskia_filters = {}
        for filter_effect in filter_.effects:
            class_ = type(filter_effect)
            if issubclass(class_, momapy.builder.Builder):
                class_ = class_._cls_to_build
            fe_func = getattr(self, self._fe_class_func_mapping[class_])
            skia_filter = fe_func(filter_effect, filter_region, dskia_filters)
            if filter_effect.result is not None:
                dskia_filters[filter_effect.result] = skia_filter
        skia_paint = skia.Paint(AntiAlias=True, ImageFilter=skia_filter)
        return skia_paint

    def _make_crop_rect_from_filter_region(self, filter_region):
        crop_rect = skia.IRect.MakeXYWH(
            round(filter_region.north_west().x),
            round(filter_region.north_west().y),
            round(filter_region.width),
            round(filter_region.height),
        )
        return crop_rect

    def _make_input_filter_from_reference(self, dskia_filters, filter_reference):
        if isinstance(filter_reference, momapy.drawing.FilterEffectInput):
            return None  # all SVG options default to source bitmap in skia
        in_skia_filter = dskia_filters.get(filter_reference)
        if in_skia_filter is None:  # if no reference or bad reference
            if dskia_filters:  # we take the last filter effect primitive if it exists, otherwise remains None (source Bitmap)
                in_skia_filter = dskia_filters[list(dskia_filters.keys())[-1]]
        return in_skia_filter

    def _make_drop_shadow_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        skia_filter = skia.ImageFilters.DropShadow(
            dx=filter_effect.dx,
            dy=filter_effect.dy,
            sigmaX=filter_effect.std_deviation,
            sigmaY=filter_effect.std_deviation,
            color=skia.Color4f(
                *filter_effect.flood_color.to_rgb(rgb_range=(0, 1)),
                filter_effect.flood_opacity,
            ),
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_composite_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        in_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in_
        )
        in2_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in2
        )
        blend_mode = self._fe_composite_comp_op_blendmode_mapping[
            filter_effect.operator
        ]  # TODO: arithmetic operator
        skia_filter = skia.ImageFilters.Xfermode(
            mode=blend_mode,
            background=in2_skia_filter,
            foreground=in_skia_filter,
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_flood_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        skia_paint = skia.Paint(
            AntiAlias=True,
            Color4f=skia.Color4f(
                *filter_effect.flood_color.to_rgb(rgb_range=(0, 1)),
                filter_effect.flood_opacity,
            ),
            Style=skia.Paint.kFill_Style,
        )
        skia_filter = skia.ImageFilters.Paint(
            paint=skia_paint,
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_gaussian_blur_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        in_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in_
        )
        tile_mode = self._fe_gaussian_blur_edgemode_tilemode_mapping[
            filter_effect.edge_mode
        ]
        skia_filter = skia.ImageFilters.Blur(
            sigmaX=filter_effect.std_deviation,
            sigmaY=filter_effect.std_deviation,
            tileMode=tile_mode,
            input=in_skia_filter,
            cropRect=crop_rect,
        )
        return skia_filter

    def _make_offset_effect(self, filter_effect, filter_region, dskia_filters):
        crop_rect = self._make_crop_rect_from_filter_region(filter_region)
        in_skia_filter = self._make_input_filter_from_reference(
            dskia_filters, filter_effect.in_
        )
        skia_filter = skia.ImageFilters.Offset(
            dx=filter_effect.dx,
            dy=filter_effect.dy,
            input=in_skia_filter,
            cropRect=crop_rect,
        )
        return skia_filter

    def _add_transform_from_drawing_element(self, drawing_element):
        if (
            drawing_element.transform is not None
            and drawing_element.transform != momapy.drawing.NoneValue
        ):
            for transformation in drawing_element.transform:
                self._add_transformation(transformation)

    def _add_transformation(self, transformation):
        class_ = type(transformation)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        tr_func = getattr(self, self._tr_class_func_mapping[class_])
        return tr_func(transformation)

    def _render_group(self, group):
        for drawing_element in group.elements:
            self.render_drawing_element(drawing_element)

    def _add_path_action_to_skia_path(self, skia_path, path_action):
        class_ = type(path_action)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        pa_func = getattr(self, self._pa_class_func_mapping[class_])
        pa_func(skia_path, path_action)

    def _make_skia_path(self, path):
        skia_path = skia.Path()
        for action in path.actions:
            self._add_path_action_to_skia_path(skia_path, action)
        return skia_path

    def _render_path(self, path):
        skia_path = self._make_skia_path(path)
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawPath(path=skia_path, paint=skia_paint)
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawPath(path=skia_path, paint=skia_paint)

    def _render_text(self, text):
        font_family = self.get_current_value("font_family")
        font_weight = self.get_current_value("font_weight")
        font_style = self.get_current_value("font_style")
        skia_typeface = self._skia_typefaces.get((font_family, font_weight, font_style))
        if skia_typeface is None:
            skia_font_slant = self._te_font_style_slant_mapping[font_style]
            skia_font_style = skia.FontStyle(
                weight=int(font_weight),
                slant=skia_font_slant,
                width=skia.FontStyle.kNormal_Width,
            )
            skia_typeface = skia.Typeface(
                familyName=font_family,
                fontStyle=skia_font_style,
            )
            self._skia_typefaces[(font_family, font_weight, font_style)] = skia_typeface
        font_size = self.get_current_value("font_size")
        skia_font = self._skia_fonts.get(
            (
                font_family,
                font_weight,
                font_style,
                font_size,
            )
        )
        if skia_font is None:
            skia_font = skia.Font(
                typeface=skia_typeface,
                size=font_size,
            )
            self._skia_fonts[
                (
                    font_family,
                    font_weight,
                    font_style,
                    font_size,
                )
            ] = skia_font
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawString(
                text=text.text,
                x=text.x,
                y=text.y,
                font=skia_font,
                paint=skia_paint,
            )
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawString(
                text=text.text,
                x=text.x,
                y=text.y,
                font=skia_font,
                paint=skia_paint,
            )

    def _render_ellipse(self, ellipse):
        skia_rect = skia.Rect(
            ellipse.x - ellipse.rx,
            ellipse.y - ellipse.ry,
            ellipse.x + ellipse.rx,
            ellipse.y + ellipse.ry,
        )
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawOval(oval=skia_rect, paint=skia_paint)
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawOval(oval=skia_rect, paint=skia_paint)

    def _render_rectangle(self, rectangle):
        skia_rect = skia.Rect(
            rectangle.x,
            rectangle.y,
            rectangle.x + rectangle.width,
            rectangle.y + rectangle.height,
        )
        if self.get_current_value("fill") != momapy.drawing.NoneValue:
            skia_paint = self._make_fill_paint()
            self.canvas.drawRoundRect(
                rect=skia_rect,
                rx=rectangle.rx,
                ry=rectangle.ry,
                paint=skia_paint,
            )
        if self.get_current_value("stroke") != momapy.drawing.NoneValue:
            skia_paint = self._make_stroke_paint()
            self.canvas.drawRoundRect(
                rect=skia_rect,
                rx=rectangle.rx,
                ry=rectangle.ry,
                paint=skia_paint,
            )

    def _add_move_to(self, skia_path, move_to):
        skia_path.moveTo(move_to.x, move_to.y)

    def _add_line_to(self, skia_path, line_to):
        skia_path.lineTo(line_to.x, line_to.y)

    def _add_curve_to(self, skia_path, curve_to):
        skia_path.cubicTo(
            curve_to.control_point1.x,
            curve_to.control_point1.y,
            curve_to.control_point2.x,
            curve_to.control_point2.y,
            curve_to.x,
            curve_to.y,
        )

    def _add_quadratic_curve_to(self, skia_path, quadratic_curve_to):
        skia_current_point = skia_path.getPoint(skia_path.countPoints() - 1)
        current_point = momapy.geometry.Point(
            skia_current_point.fX, skia_current_point.fY
        )
        curve_to = quadratic_curve_to.to_curve_to(current_point)
        self._add_curve_to(skia_path, curve_to)

    def _add_close_path(self, skia_path, close_path):
        skia_path.close()

    def _add_elliptical_arc(self, skia_path, elliptical_arc):
        if elliptical_arc.arc_flag == 0:
            skia_arc_flag = skia.Path.ArcSize.kSmall_ArcSize
        else:
            skia_arc_flag = skia.Path.ArcSize.kLarge_ArcSize
        if elliptical_arc.sweep_flag == 1:
            skia_sweep_flag = skia.PathDirection.kCW
        else:
            skia_sweep_flag = skia.PathDirection.kCCW
        skia_path.arcTo(
            rx=elliptical_arc.rx,
            ry=elliptical_arc.ry,
            xAxisRotate=elliptical_arc.x_axis_rotation,
            largeArc=skia_arc_flag,
            sweep=skia_sweep_flag,
            x=elliptical_arc.x,
            y=elliptical_arc.y,
        )

    def _add_translation(self, translation):
        self.canvas.translate(dx=translation.tx, dy=translation.ty)

    def _add_rotation(self, rotation):
        angle = math.degrees(rotation.angle)
        if rotation.point is not None:
            self.canvas.rotate(degrees=angle, px=rotation.point.x, py=rotation.point.y)
        else:
            self.canvas.rotate(degrees=angle)

    def _add_scaling(self, scaling):
        self.canvas.scale(sx=scaling.sx, sy=scaling.sy)

    def _add_matrix_transformation(self, matrix_transformation):
        m = skia.Matrix.MakeAll(
            scaleX=matrix_transformation.m[0][0],
            skewX=matrix_transformation.m[0][1],
            transX=matrix_transformation.m[0][2],
            skewY=matrix_transformation.m[1][0],
            scaleY=matrix_transformation.m[1][1],
            transY=matrix_transformation.m[1][2],
            pers0=matrix_transformation.m[2][0],
            pers1=matrix_transformation.m[2][1],
            pers2=matrix_transformation.m[2][2],
        )
        self.canvas.concat(m)



"""Tests for CellDesigner reading functionality."""

import pytest
import os
import momapy.io.core
import momapy.celldesigner.core
import momapy.core.layout
import momapy.sbml.core
import frozendict


# Get the directory containing this test file
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
CELLDESIGNER_MAPS_DIR = os.path.join(TEST_DIR, "..", "maps")

# Discover all .xml files in the maps directory
def get_celldesigner_files():
    """Get all CellDesigner XML files from the maps directory."""
    if not os.path.exists(CELLDESIGNER_MAPS_DIR):
        return []
    return [f for f in os.listdir(CELLDESIGNER_MAPS_DIR) if f.endswith('.xml')]

CELLDESIGNER_FILES = get_celldesigner_files()


class TestCellDesignerReading:
    """Tests for reading CellDesigner files."""

    @pytest.mark.parametrize("filename", CELLDESIGNER_FILES)
    def test_read_celldesigner_file(self, filename):
        """Test reading CellDesigner XML files."""
        input_file = os.path.join(CELLDESIGNER_MAPS_DIR, filename)
        if not os.path.exists(input_file):
            pytest.skip(f"CellDesigner file {filename} not found")
        result = momapy.io.core.read(input_file)
        assert result is not None
        assert result.obj is not None
        # Verify we got a map object
        assert hasattr(result.obj, 'layout')


class TestCellDesignerReadOptionalParameters:
    """Tests for CellDesigner read function optional parameters."""

    @pytest.fixture
    def test_file(self):
        """Return path to a test CellDesigner file."""
        if not CELLDESIGNER_FILES:
            pytest.skip("No CellDesigner test files found")
        return os.path.join(CELLDESIGNER_MAPS_DIR, CELLDESIGNER_FILES[0])

    @pytest.mark.parametrize("return_type,expected_type", [
        ("map", momapy.celldesigner.core.CellDesignerMap),
        ("model", momapy.celldesigner.core.CellDesignerModel),
        ("layout", momapy.core.layout.Layout),
    ])
    def test_return_type_parameter(self, test_file, return_type, expected_type):
        """Test return_type parameter returns correct object type."""
        result = momapy.io.core.read(test_file, return_type=return_type)
        assert isinstance(result.obj, expected_type)

    def test_with_model_true(self, test_file):
        """Test with_model=True includes model in result."""
        result = momapy.io.core.read(test_file, return_type="map", with_model=True)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.celldesigner.core.CellDesignerMap)
        # Verify the map has a model
        assert hasattr(result.obj, 'model')
        assert result.obj.model is not None

    def test_with_model_false(self, test_file):
        """Test with_model=False excludes model from result."""
        result = momapy.io.core.read(test_file, return_type="map", with_model=False)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.celldesigner.core.CellDesignerMap)
        # Verify the map has no model (or model is None)
        assert hasattr(result.obj, 'model')
        assert result.obj.model is None

    def test_with_layout_true(self, test_file):
        """Test with_layout=True includes layout in result."""
        result = momapy.io.core.read(test_file, return_type="map", with_layout=True)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.celldesigner.core.CellDesignerMap)
        # Verify the map has a layout
        assert hasattr(result.obj, 'layout')
        assert result.obj.layout is not None

    def test_with_layout_false(self, test_file):
        """Test with_layout=False excludes layout from result."""
        result = momapy.io.core.read(test_file, return_type="map", with_layout=False)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.celldesigner.core.CellDesignerMap)
        # Verify the map has no layout (or layout is None)
        assert hasattr(result.obj, 'layout')
        assert result.obj.layout is None

    def test_with_annotations_true(self, test_file):
        """Test with_annotations=True includes annotations in result."""
        result = momapy.io.core.read(test_file, with_annotations=True)
        assert hasattr(result, 'annotations')
        assert isinstance(result.annotations, frozendict.frozendict)
        # Annotations may be empty if file has none, but should be a frozendict

    def test_with_annotations_false(self, test_file):
        """Test with_annotations=False excludes annotations from result."""
        result = momapy.io.core.read(test_file, with_annotations=False)
        assert hasattr(result, 'annotations')
        # Should be empty frozendict when with_annotations=False
        assert result.annotations == frozendict.frozendict()

    def test_with_notes_true(self, test_file):
        """Test with_notes=True includes notes in result."""
        result = momapy.io.core.read(test_file, with_notes=True)
        assert hasattr(result, 'notes')
        assert isinstance(result.notes, frozendict.frozendict)
        # Notes may be empty if file has none, but should be a frozendict

    def test_with_notes_false(self, test_file):
        """Test with_notes=False excludes notes from result."""
        result = momapy.io.core.read(test_file, with_notes=False)
        assert hasattr(result, 'notes')
        # Should be empty frozendict when with_notes=False
        assert result.notes == frozendict.frozendict()


class TestCellDesignerAnnotationsContent:
    """Tests that annotations are correctly parsed from CellDesigner files.

    Uses Apoptosis_pathway.xml which contains RDF annotations on compartments,
    species, and reactions.
    """

    APOPTOSIS = os.path.join(CELLDESIGNER_MAPS_DIR, "Apoptosis_pathway.xml")

    @pytest.fixture(scope="class")
    def result(self):
        if not os.path.exists(self.APOPTOSIS):
            pytest.skip("Apoptosis_pathway.xml not found")
        return momapy.io.core.read(self.APOPTOSIS, with_annotations=True, with_notes=True)

    def _get_element_by_id(self, result, id_):
        """Find an annotated element by its id_."""
        for elem in result.annotations:
            if getattr(elem, "id_", None) == id_:
                return elem
        return None

    def _get_annotations_by_id(self, result, id_):
        """Get the frozenset of annotations for an element by its id_."""
        elem = self._get_element_by_id(result, id_)
        if elem is not None:
            return result.annotations.get(elem, frozenset())
        return frozenset()

    def test_annotations_are_non_empty(self, result):
        """Test that the file produces a non-empty annotations dict."""
        non_empty = {
            elem: annots
            for elem, annots in result.annotations.items()
            if annots
        }
        assert len(non_empty) > 0

    def test_annotation_values_are_frozensets_of_rdf_annotations(self, result):
        """Test that each value is a frozenset of RDFAnnotation objects."""
        for elem, annots in result.annotations.items():
            assert isinstance(annots, frozenset)
            for a in annots:
                assert isinstance(a, momapy.sbml.core.RDFAnnotation)

    def test_compartment_annotation(self, result):
        """Test annotation on compartment s_id_ca4 (plasma membrane)."""
        annots = self._get_annotations_by_id(result, "s_id_ca4")
        assert len(annots) == 1
        annotation = next(iter(annots))
        assert annotation.qualifier == momapy.sbml.core.BQBiol.IS
        assert annotation.resources == frozenset(
            {"urn:miriam:obo.go:GO%3A0005886"}
        )

    def test_species_annotations(self, result):
        """Test annotations on species s_id_sa30 (BAX protein)."""
        annots = self._get_annotations_by_id(result, "s_id_sa30")
        assert len(annots) == 6
        qualifiers = {a.qualifier for a in annots}
        assert qualifiers == {momapy.sbml.core.BQBiol.IS_DESCRIBED_BY}
        resources = {r for a in annots for r in a.resources}
        assert "urn:miriam:uniprot:Q07812" in resources
        assert "urn:miriam:hgnc.symbol:BAX" in resources
        assert "urn:miriam:hgnc:959" in resources

    def test_annotations_span_multiple_element_types(self, result):
        """Test that annotations cover compartments, species, and reactions."""
        types_with_annotations = {
            type(elem).__name__
            for elem, annots in result.annotations.items()
            if annots
        }
        assert "Compartment" in types_with_annotations
        assert "GenericProtein" in types_with_annotations
        assert "StateTransition" in types_with_annotations


class TestCellDesignerNotesContent:
    """Tests that notes are correctly parsed from CellDesigner files.

    Uses Apoptosis_pathway.xml which contains notes on compartments,
    species, and the map itself.
    """

    APOPTOSIS = os.path.join(CELLDESIGNER_MAPS_DIR, "Apoptosis_pathway.xml")

    @pytest.fixture(scope="class")
    def result(self):
        if not os.path.exists(self.APOPTOSIS):
            pytest.skip("Apoptosis_pathway.xml not found")
        return momapy.io.core.read(self.APOPTOSIS, with_annotations=True, with_notes=True)

    def test_notes_are_non_empty(self, result):
        """Test that the file produces a non-empty notes dict."""
        non_empty = {
            elem: notes
            for elem, notes in result.notes.items()
            if notes
        }
        assert len(non_empty) > 0

    def test_notes_values_are_frozensets_of_strings(self, result):
        """Test that each value is a frozenset of str."""
        for elem, notes in result.notes.items():
            assert isinstance(notes, frozenset)
            for n in notes:
                assert isinstance(n, str)

    def test_notes_contain_xml(self, result):
        """Test that note strings are valid XML fragments."""
        for elem, notes in result.notes.items():
            for n in notes:
                assert n.startswith("<")
                break
            break

    def test_map_level_notes(self, result):
        """Test that the map object has notes."""
        map_notes = result.notes.get(result.obj)
        assert map_notes is not None
        assert len(map_notes) == 1
        note = next(iter(map_notes))
        assert isinstance(note, str)
        assert "<html" in note

    def test_compartment_notes_contain_description(self, result):
        """Test that compartment s_id_ca4 (plasma membrane) has notes."""
        for elem, notes in result.notes.items():
            if getattr(elem, "id_", None) == "s_id_ca4" and notes:
                note = next(iter(notes))
                assert "plasma membrane" in note
                return
        pytest.fail("No notes found for compartment s_id_ca4")

    def test_notes_span_multiple_element_types(self, result):
        """Test that notes cover compartments, species, and the map."""
        types_with_notes = {
            type(elem).__name__
            for elem, notes in result.notes.items()
            if notes
        }
        assert "Compartment" in types_with_notes
        assert "GenericProtein" in types_with_notes
        assert "CellDesignerMap" in types_with_notes


class TestCellDesignerEmptyModifications:
    """Test that modification residues absent from listOfModifications are
    stored as Modification(state=None), matching the state="empty" behaviour.

    Uses Neuroinflammation.xml which contains MAPK14 (protein p_7, residues
    T180/mr_re_1_1 and Y182/mr_re_2_1):

    - s_id_ne10f: empty <listOfModifications> → both residues should appear
      with state=None after the fix.
    - s_id_ne207: mr_re_1_1 listed as phosphorylated → mr_re_2_1 should be
      added with state=None by the fix.
    """

    NEUROINFLAMMATION = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "maps", "Neuroinflammation.xml"
    )

    @pytest.fixture(scope="class")
    def neuroinflammation_map(self):
        if not os.path.exists(self.NEUROINFLAMMATION):
            pytest.skip("Neuroinflammation.xml not found")
        return momapy.io.core.read(self.NEUROINFLAMMATION).obj

    def _get_species(self, map_, species_id):
        return next(s for s in map_.model.species if s.id_ == species_id)

    def test_empty_list_of_modifications_creates_residues_from_template(
        self, neuroinflammation_map
    ):
        # s_id_ne10f has an empty <listOfModifications>; p_7 template defines
        # two residues, so both must appear with state=None.
        species = self._get_species(neuroinflammation_map, "s_id_ne10f")
        assert len(species.modifications) == 2
        for mod in species.modifications:
            assert mod.state is None

    def test_partial_list_of_modifications_fills_uncovered_residues(
        self, neuroinflammation_map
    ):
        # s_id_ne207 has mr_re_1_1 explicitly listed as phosphorylated.
        # The fix must add mr_re_2_1 with state=None.
        species = self._get_species(neuroinflammation_map, "s_id_ne207")
        assert len(species.modifications) == 2
        states = {mod.state for mod in species.modifications}
        assert momapy.celldesigner.core.ModificationState.PHOSPHORYLATED in states
        assert None in states



"""Library for working with molecular maps programmatically.

This package provides tools for reading, writing, manipulating, and rendering
molecular maps, with support for various formats including SBGN-ML and CellDesigner.

Example:
    >>> from momapy.io.core import read
    >>> from momapy.rendering.core import render_layout_elements
    >>> # Read a molecular map
    >>> result = read("map.sbgn", return_type="layout")
    >>> layout = result.obj
    >>> # Render the map to an image
    >>> render_layout_elements(
    ...     layout_elements=[layout],
    ...     file_path="output.svg",
    ...     format_="svg"
    ... )
"""



"""SBGN-ML writer classes."""

import dataclasses
import typing

import lxml.etree

import momapy.core.elements
import momapy.core.layout
import momapy.io.core
import momapy.sbgn.core
import momapy.sbgn.pd
import momapy.sbgn.af
import momapy.sbml.io.sbml._qualifiers


@dataclasses.dataclass
class WritingContext:
    """Bundles the shared state passed across all writer methods."""

    map_: typing.Any
    annotations: dict
    notes: dict
    ids: dict
    with_annotations: bool
    with_notes: bool


class _SBGNMLWriter(momapy.io.core.Writer):
    _NSMAP = {
        None: "http://sbgn.org/libsbgn/0.3",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "bqmodel": "http://biomodels.net/model-qualifiers/",
        "bqbiol": "http://biomodels.net/biology-qualifiers/",
    }
    _DIRECTION_TO_SBGNML_ORIENTATION = {
        momapy.core.elements.Direction.HORIZONTAL: "horizontal",
        momapy.core.elements.Direction.VERTICAL: "vertical",
        momapy.core.elements.Direction.RIGHT: "right",
        momapy.core.elements.Direction.LEFT: "left",
        momapy.core.elements.Direction.DOWN: "down",
        momapy.core.elements.Direction.UP: "up",
    }
    _CLASS_TO_SBGNML_CLASS_ATTRIBUTE = {
        momapy.sbgn.pd.SBGNPDMap: "process description",
        momapy.sbgn.pd.SBGNPDModel: "process description",
        momapy.sbgn.pd.SBGNPDLayout: "process description",
        momapy.sbgn.pd.StateVariableLayout: "state variable",
        momapy.sbgn.pd.UnitOfInformationLayout: "unit of information",
        momapy.sbgn.pd.TerminalLayout: "terminal",
        momapy.sbgn.pd.MacromoleculeSubunitLayout: "macromolecule",
        momapy.sbgn.pd.SimpleChemicalSubunitLayout: "simple chemical",
        momapy.sbgn.pd.NucleicAcidFeatureSubunitLayout: "nucleic acid feature",
        momapy.sbgn.pd.ComplexSubunitLayout: "complex",
        momapy.sbgn.pd.MacromoleculeMultimerSubunitLayout: "macromolecule multimer",
        momapy.sbgn.pd.SimpleChemicalMultimerSubunitLayout: "simple chemical multimer",
        momapy.sbgn.pd.NucleicAcidFeatureMultimerSubunitLayout: "nucleic acid feature multimer",
        momapy.sbgn.pd.ComplexMultimerSubunitLayout: "complex multimer",
        momapy.sbgn.pd.CompartmentLayout: "compartment",
        momapy.sbgn.pd.SubmapLayout: "submap",
        momapy.sbgn.pd.UnspecifiedEntityLayout: "unspecified entity",
        momapy.sbgn.pd.MacromoleculeLayout: "macromolecule",
        momapy.sbgn.pd.SimpleChemicalLayout: "simple chemical",
        momapy.sbgn.pd.NucleicAcidFeatureLayout: "nucleic acid feature",
        momapy.sbgn.pd.ComplexLayout: "complex",
        momapy.sbgn.pd.MacromoleculeMultimerLayout: "macromolecule multimer",
        momapy.sbgn.pd.SimpleChemicalMultimerLayout: "simple chemical multimer",
        momapy.sbgn.pd.NucleicAcidFeatureMultimerLayout: "nucleic acid feature multimer",
        momapy.sbgn.pd.ComplexMultimerLayout: "complex multimer",
        momapy.sbgn.pd.PerturbingAgentLayout: "perturbing agent",
        momapy.sbgn.pd.EmptySetLayout: "empty set",
        momapy.sbgn.pd.TagLayout: "tag",
        momapy.sbgn.pd.GenericProcessLayout: "process",
        momapy.sbgn.pd.UncertainProcessLayout: "uncertain process",
        momapy.sbgn.pd.OmittedProcessLayout: "omitted process",
        momapy.sbgn.pd.AssociationLayout: "association",
        momapy.sbgn.pd.DissociationLayout: "dissociation",
        momapy.sbgn.pd.PhenotypeLayout: "phenotype",
        momapy.sbgn.pd.AndOperatorLayout: "and",
        momapy.sbgn.pd.OrOperatorLayout: "or",
        momapy.sbgn.pd.NotOperatorLayout: "not",
        momapy.sbgn.pd.EquivalenceOperatorLayout: "equivalence",
        momapy.sbgn.pd.ConsumptionLayout: "consumption",
        momapy.sbgn.pd.ProductionLayout: "production",
        momapy.sbgn.pd.ModulationLayout: "modulation",
        momapy.sbgn.pd.StimulationLayout: "stimulation",
        momapy.sbgn.pd.CatalysisLayout: "catalysis",
        momapy.sbgn.pd.NecessaryStimulationLayout: "necessary stimulation",
        momapy.sbgn.pd.InhibitionLayout: "inhibition",
        momapy.sbgn.pd.LogicArcLayout: "logic arc",
        momapy.sbgn.pd.EquivalenceArcLayout: "equivalence arc",
        momapy.sbgn.af.CompartmentLayout: "compartment",
        momapy.sbgn.af.SubmapLayout: "submap",
        momapy.sbgn.af.BiologicalActivityLayout: "biological activity",
        momapy.sbgn.af.UnspecifiedEntityUnitOfInformationLayout: "unspecified entity",
        momapy.sbgn.af.MacromoleculeUnitOfInformationLayout: "macromolecule",
        momapy.sbgn.af.SimpleChemicalUnitOfInformationLayout: "simple chemical",
        momapy.sbgn.af.NucleicAcidFeatureUnitOfInformationLayout: "nucleic acid feature",
        momapy.sbgn.af.ComplexUnitOfInformationLayout: "complex",
        momapy.sbgn.af.PerturbationUnitOfInformationLayout: "perturbation",
        momapy.sbgn.af.PhenotypeLayout: "phenotype",
        momapy.sbgn.af.AndOperatorLayout: "and",
        momapy.sbgn.af.OrOperatorLayout: "or",
        momapy.sbgn.af.NotOperatorLayout: "not",
        momapy.sbgn.af.DelayOperatorLayout: "delay",
        momapy.sbgn.af.UnknownInfluenceLayout: "unknown influence",
        momapy.sbgn.af.PositiveInfluenceLayout: "positive influence",
        momapy.sbgn.af.NecessaryStimulationLayout: "necessary stimulation",
        momapy.sbgn.af.NegativeInfluenceLayout: "negative influence",
        momapy.sbgn.af.TerminalLayout: "terminal",
        momapy.sbgn.af.TagLayout: "tag",
        momapy.sbgn.af.LogicArcLayout: "logic arc",
        momapy.sbgn.af.EquivalenceArcLayout: "equivalence arc",
    }

    @classmethod
    def _make_lxml_element(
        cls, tag, namespace=None, attributes=None, text=None, nsmap=None
    ):
        if namespace is not None:
            lxml_tag = f"{{{namespace}}}{tag}"
        else:
            lxml_tag = tag
        if nsmap is None:
            nsmap = {}
        if attributes is None:
            attributes = {}
        lxml_element = lxml.etree.Element(lxml_tag, nsmap=nsmap, **attributes)
        if text is not None:
            lxml_element.text = text
        return lxml_element

    @classmethod
    def _get_sbgnml_id_from_map_element(cls, map_element, ids):
        sbgnml_ids = ids.get(map_element)
        if sbgnml_ids is None:
            sbgnml_id = map_element.id_
        else:
            sbgnml_id = sbgnml_ids[0]
        return sbgnml_id

    @classmethod
    def write(
        cls,
        obj: momapy.sbgn.core.SBGNMap,
        file_path,
        annotations=None,
        notes=None,
        ids=None,
        with_render_information=True,
        with_annotations=True,
        with_notes=True,
    ):
        if annotations is None:
            annotations = {}
        if notes is None:
            notes = {}
        if ids is None:
            ids = {}
        ctx = WritingContext(
            map_=obj,
            annotations=annotations,
            notes=notes,
            ids=ids,
            with_annotations=with_annotations,
            with_notes=with_notes,
        )
        sbgnml_sbgn = cls._make_lxml_element("sbgn", nsmap=cls._NSMAP)
        sbgnml_map = cls._make_sbgnml_map_from_map(ctx)
        sbgnml_sbgn.append(sbgnml_map)
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(
                lxml.etree.tostring(
                    sbgnml_sbgn, pretty_print=True, xml_declaration=True
                ).decode()
            )

    @classmethod
    def _make_sbgnml_map_from_map(cls, ctx):
        language = cls._CLASS_TO_SBGNML_CLASS_ATTRIBUTE[type(ctx.map_)]
        id_ = ctx.ids.get(ctx.map_)
        if id_ is None:
            id_ = ctx.map_.id_
        else:
            id_ = id_[0]
        attributes = {"id": id_, "language": language}
        sbgnml_map = cls._make_lxml_element("map", attributes=attributes)
        sbgnml_bbox = cls._make_sbgnml_bbox_from_node(ctx.map_.layout)
        sbgnml_map.append(sbgnml_bbox)
        for layout_element in ctx.map_.layout.layout_elements:
            cls._make_and_add_sbgnml_elements_from_layout_element(
                ctx=ctx,
                sbgnml_map=sbgnml_map,
                layout_element=layout_element,
            )
        return sbgnml_map

    @classmethod
    def _make_and_add_sbgnml_elements_from_layout_element(
        cls, ctx, sbgnml_map, layout_element
    ):
        model_element = ctx.map_.get_mapping(layout_element)
        if model_element is not None:
            if isinstance(layout_element, momapy.core.layout.Node):
                sbgnml_elements = cls._make_sbgnml_elements_from_node(
                    ctx=ctx,
                    node=layout_element,
                )
            elif isinstance(
                layout_element,
                (momapy.core.layout.SingleHeadedArc, momapy.core.layout.DoubleHeadedArc),
            ):
                sbgnml_elements = cls._make_sbgnml_elements_from_arc(
                    ctx=ctx,
                    arc=layout_element,
                )
            if ctx.with_annotations:
                if ctx.annotations is not None:
                    element_annotations = ctx.annotations.get(model_element)
                    if element_annotations is not None:
                        sbgnml_annotation = (
                            cls._make_sbgnml_annotation_from_annotations(
                                element_annotations,
                                sbgnml_id=cls._get_sbgnml_id_from_map_element(
                                    model_element, ctx.ids
                                ),
                            )
                        )
                        sbgnml_extension = cls._make_lxml_element("extension")
                        sbgnml_extension.append(sbgnml_annotation)
                        for sbgnml_element in sbgnml_elements:
                            sbgnml_element.append(sbgnml_extension)
            if ctx.with_notes:
                if ctx.notes is not None:
                    element_notes = ctx.notes.get(model_element)
                    if element_notes is not None:
                        for note in element_notes:
                            sbgnml_notes = cls._make_lxml_element(tag="notes")
                            notes_root = lxml.etree.fromstring(note)
                            sbgnml_notes.append(notes_root)
                            for sbgnml_element in sbgnml_elements:
                                sbgnml_element.append(sbgnml_notes)
            for sbgnml_element in sbgnml_elements:
                sbgnml_map.append(sbgnml_element)

    @classmethod
    def _make_sbgnml_annotation_from_annotations(cls, annotations, sbgnml_id):
        sbgnml_annotation = cls._make_lxml_element("annotation")
        sbgnml_rdf = cls._make_lxml_element(
            tag="RDF", namespace=cls._NSMAP["rdf"], nsmap=cls._NSMAP
        )
        sbgnml_annotation.append(sbgnml_rdf)
        sbgnml_description = cls._make_lxml_element(
            tag="Description",
            namespace=cls._NSMAP["rdf"],
            attributes={f"{{{cls._NSMAP['rdf']}}}about": f"#{sbgnml_id}"},
        )
        sbgnml_rdf.append(sbgnml_description)
        for annotation in annotations:
            namespace, tag = momapy.sbml.io.sbml._qualifiers.QUALIFIER_MEMBER_TO_QUALIFIER_ATTRIBUTE[
                annotation.qualifier
            ]
            sbgnml_bq = cls._make_lxml_element(tag=tag, namespace=namespace)
            sbgnml_description.append(sbgnml_bq)
            sbgnml_bag = cls._make_lxml_element(tag="Bag", namespace=cls._NSMAP["rdf"])
            sbgnml_bq.append(sbgnml_bag)
            for resource in annotation.resources:
                sbgnml_li = cls._make_lxml_element(
                    tag="li",
                    namespace=cls._NSMAP["rdf"],
                    attributes={f"{{{cls._NSMAP['rdf']}}}resource": resource},
                )
                sbgnml_bag.append(sbgnml_li)
        return sbgnml_annotation

    @classmethod
    def _make_sbgnml_elements_from_node(cls, ctx, node):
        sbgnml_id = cls._get_sbgnml_id_from_map_element(node, ctx.ids)
        sbgnml_class = cls._CLASS_TO_SBGNML_CLASS_ATTRIBUTE[type(node)]
        attributes = {"id": sbgnml_id, "class": sbgnml_class}
        direction = getattr(node, "direction", None)
        if direction is not None:
            sbgnml_orientation = cls._DIRECTION_TO_SBGNML_ORIENTATION[direction]
            attributes["orientation"] = sbgnml_orientation
        model_element = ctx.map_.get_mapping(node)
        if isinstance(
            model_element, (momapy.sbgn.pd.EntityPool, momapy.sbgn.af.Activity)
        ):
            compartment = model_element.compartment
            if compartment is not None:
                compartment_id = cls._get_sbgnml_id_from_map_element(compartment, ctx.ids)
                attributes["compartmentRef"] = compartment_id
        sbgnml_element = cls._make_lxml_element("glyph", attributes=attributes)
        sbgnml_elements = [sbgnml_element]
        sbgnml_bbox = cls._make_sbgnml_bbox_from_node(node)
        sbgnml_element.append(sbgnml_bbox)
        if node.label is not None:
            if isinstance(node, momapy.sbgn.pd.StateVariableLayout):
                sbgnml_state = cls._make_sbgnml_state_from_text_layout(node.label)
                sbgnml_element.append(sbgnml_state)
            else:
                sbgnml_label = cls._make_sbgnml_label_from_text_layout(node.label)
                sbgnml_element.append(sbgnml_label)
        for side, attr in [("left", "left_connector_tip"), ("right", "right_connector_tip")]:
            if hasattr(node, attr):
                connector_tip = getattr(node, attr)()
                sbgnml_port = cls._make_sbgnml_port_from_point(
                    connector_tip, port_id=f"{sbgnml_id}_{side}"
                )
                sbgnml_element.append(sbgnml_port)
        for layout_element in node.layout_elements:
            if isinstance(layout_element, momapy.core.layout.Node):
                sub_sbgnml_elements = cls._make_sbgnml_elements_from_node(
                    ctx=ctx,
                    node=layout_element,
                )
                for sub_sbgnml_element in sub_sbgnml_elements:
                    if sub_sbgnml_element.tag == "glyph":
                        sbgnml_element.append(sub_sbgnml_element)
                    else:
                        sbgnml_elements.append(sub_sbgnml_element)
        return sbgnml_elements

    @classmethod
    def _make_sbgnml_elements_from_arc(cls, ctx, arc):
        sbgnml_id = cls._get_sbgnml_id_from_map_element(arc, ctx.ids)
        sbgnml_class = cls._CLASS_TO_SBGNML_CLASS_ATTRIBUTE[type(arc)]
        attributes = {
            "id": sbgnml_id,
            "class": sbgnml_class,
        }
        points = arc.points()
        sbgnml_source_id = cls._get_sbgnml_id_from_map_element(arc.source, ctx.ids)
        sbgnml_target_id = cls._get_sbgnml_id_from_map_element(arc.target, ctx.ids)
        # momapy reverts the consumption and logic arc direction compared to
        # SBGN-ML, so we need to revert it back here
        if isinstance(
            arc,
            (
                momapy.sbgn.pd.ConsumptionLayout,
                momapy.sbgn.pd.LogicArcLayout,
                momapy.sbgn.af.LogicArcLayout,
                momapy.sbgn.pd.EquivalenceArcLayout,
                momapy.sbgn.af.EquivalenceArcLayout,
            ),
        ):
            attributes["source"] = sbgnml_target_id
            attributes["target"] = sbgnml_source_id
            points.reverse()
        else:
            attributes["target"] = sbgnml_target_id
            attributes["source"] = sbgnml_source_id
        sbgnml_element = cls._make_lxml_element("arc", attributes=attributes)
        sbgnml_points = cls._make_sbgnml_points_from_points(points)
        for sbgnml_point in sbgnml_points:
            sbgnml_element.append(sbgnml_point)
        return [sbgnml_element]

    @classmethod
    def _make_sbgnml_points_from_points(cls, points):
        sbgnml_elements = []
        start_point = points[0]
        sbgnml_start_point_attributes = {
            "x": str(start_point.x),
            "y": str(start_point.y),
        }
        sbgnml_start_point = cls._make_lxml_element(
            "start", attributes=sbgnml_start_point_attributes
        )
        sbgnml_elements.append(sbgnml_start_point)
        for point in points[1:-1]:
            sbgnml_next_point_attributes = {
                "x": str(point.x),
                "y": str(point.y),
            }
            sbgnml_next_point = cls._make_lxml_element(
                "next", attributes=sbgnml_next_point_attributes
            )
            sbgnml_elements.append(sbgnml_next_point)
        end_point = points[-1]
        sbgnml_end_point_attributes = {
            "x": str(end_point.x),
            "y": str(end_point.y),
        }
        sbgnml_end_point = cls._make_lxml_element(
            "end", attributes=sbgnml_end_point_attributes
        )
        sbgnml_elements.append(sbgnml_end_point)
        return sbgnml_elements

    @classmethod
    def _make_sbgnml_port_from_point(cls, point, port_id):
        attributes = {"id": port_id, "x": str(point.x), "y": str(point.y)}
        sbgnml_element = cls._make_lxml_element("port", attributes=attributes)
        return sbgnml_element

    @classmethod
    def _make_sbgnml_bbox_from_node(cls, node):
        attributes = {
            "x": str(node.x - node.width / 2),
            "y": str(node.y - node.height / 2),
            "w": str(node.width),
            "h": str(node.height),
        }
        sbgnml_bbox = cls._make_lxml_element("bbox", attributes=attributes)
        return sbgnml_bbox

    @classmethod
    def _make_sbgnml_bbox_from_text_layout(cls, text_layout):
        bbox = text_layout.bbox()
        attributes = {
            "x": str(bbox.x - bbox.width / 2),
            "y": str(bbox.y - bbox.height / 2),
            "w": str(bbox.width),
            "h": str(bbox.height),
        }
        sbgnml_bbox = cls._make_lxml_element("bbox", attributes=attributes)
        return sbgnml_bbox

    @classmethod
    def _make_sbgnml_label_from_text_layout(cls, text_layout):
        attributes = {"text": text_layout.text}
        sbgnml_label = cls._make_lxml_element("label", attributes=attributes)
        sbgnml_bbox = cls._make_sbgnml_bbox_from_text_layout(text_layout)
        sbgnml_label.append(sbgnml_bbox)
        return sbgnml_label

    @classmethod
    def _make_sbgnml_state_from_text_layout(cls, text_layout):
        attributes = {}
        text_split = text_layout.text.split("@")
        if len(text_split) > 1:
            attributes["variable"] = text_split[-1]
        if text_split[0]:
            attributes["value"] = text_split[0]
        sbgnml_state = cls._make_lxml_element("state", attributes=attributes)
        return sbgnml_state


class SBGNML0_3Writer(_SBGNMLWriter):
    """Class for SBGN-ML 0.3 writer objects"""

    pass



"""Functions for positioning layout elements and related objects relatively to other objects.

This module provides utilities for calculating positions relative to layout elements,
fitting bounding boxes around collections of elements, and setting positions of
builder objects. Functions support Points, Bboxes, Nodes, and their builder variants.

Example:
    >>> from momapy.positioning import right_of, fit
    >>> from momapy.geometry import Point
    >>> point = Point(0, 0)
    >>> new_point = right_of(point, 10)
    >>> print(new_point)
    Point(x=10, y=0)
"""

import collections.abc

import momapy.core
import momapy.core.builders
import momapy.core.elements
import momapy.core.layout
import momapy.geometry
import momapy.builder


def right_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Return a point to the right of the given object at the specified distance.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the east anchor point.
        distance: The horizontal distance from the object's reference point.

    Returns:
        A new Point located to the right of the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(0, 0)
        >>> right_of(point, 10)
        Point(x=10, y=0)
    """
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.east()
    else:
        raise TypeError
    return source_point + (distance, 0)


def left_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Return a point to the left of the given object at the specified distance.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the west anchor point.
        distance: The horizontal distance from the object's reference point.

    Returns:
        A new Point located to the left of the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(10, 0)
        >>> left_of(point, 5)
        Point(x=5, y=0)
    """
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.west()
    else:
        raise TypeError
    return source_point - (distance, 0)


def above_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Return a point above the given object at the specified distance.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the north anchor point.
        distance: The vertical distance from the object's reference point.

    Returns:
        A new Point located above the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(0, 10)
        >>> above_of(point, 5)
        Point(x=0, y=5)
    """
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.north()
    else:
        raise TypeError
    return source_point - (0, distance)


def below_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
) -> momapy.geometry.Point:
    """Return a point below the given object at the specified distance.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the south anchor point.
        distance: The vertical distance from the object's reference point.

    Returns:
        A new Point located below the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(0, 0)
        >>> below_of(point, 10)
        Point(x=0, y=10)
    """
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.south()
    else:
        raise TypeError
    return source_point + (0, distance)


def above_left_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Return a point above and to the left of the given object.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the north-west anchor point.
        distance1: The vertical distance (northward) from the object.
        distance2: The horizontal distance (westward) from the object.
            If None, defaults to distance1.

    Returns:
        A new Point located above and to the left of the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(10, 10)
        >>> above_left_of(point, 5)
        Point(x=5, y=5)
    """
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.north_west()
    else:
        raise TypeError
    return source_point - (distance2, distance1)


def above_right_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Return a point above and to the right of the given object.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the north-east anchor point.
        distance1: The vertical distance (northward) from the object.
        distance2: The horizontal distance (eastward) from the object.
            If None, defaults to distance1.

    Returns:
        A new Point located above and to the right of the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(0, 10)
        >>> above_right_of(point, 5, 10)
        Point(x=10, y=5)
    """
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.north_east()
    else:
        raise TypeError
    return source_point + (distance2, -distance1)


def below_left_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Return a point below and to the left of the given object.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the south-west anchor point.
        distance1: The vertical distance (southward) from the object.
        distance2: The horizontal distance (westward) from the object.
            If None, defaults to distance1.

    Returns:
        A new Point located below and to the left of the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(10, 0)
        >>> below_left_of(point, 5)
        Point(x=5, y=5)
    """
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.south_west()
    else:
        raise TypeError
    return source_point + (-distance2, distance1)


def below_right_of(
    obj: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
) -> momapy.geometry.Point:
    """Return a point below and to the right of the given object.

    Args:
        obj: A Point, Bbox, LayoutElement, or their builder variants.
            For Nodes, uses the south-east anchor point.
        distance1: The vertical distance (southward) from the object.
        distance2: The horizontal distance (eastward) from the object.
            If None, defaults to distance1.

    Returns:
        A new Point located below and to the right of the object.

    Raises:
        TypeError: If obj is not a supported type.

    Example:
        >>> point = momapy.geometry.Point(0, 0)
        >>> below_right_of(point, 5, 10)
        Point(x=10, y=5)
    """
    if distance2 is None:
        distance2 = distance1
    if momapy.builder.isinstance_or_builder(obj, momapy.geometry.Point):
        source_point = obj
    elif momapy.builder.isinstance_or_builder(obj, momapy.core.layout.Node):
        source_point = obj.south_east()
    else:
        raise TypeError
    return source_point + (distance2, distance1)


def fit(
    elements: collections.abc.Collection[
        momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.geometry.Point
        | momapy.geometry.PointBuilder
    ],
    xsep: float = 0,
    ysep: float = 0,
) -> momapy.geometry.Bbox:
    """Compute a bounding box that fits the given elements with optional margins.

    Args:
        elements: A collection of Points, Bboxes, LayoutElements, or their
            builder variants to fit within the bounding box.
        xsep: Horizontal margin to add to both sides of the bounding box.
        ysep: Vertical margin to add to both sides of the bounding box.

    Returns:
        A Bbox that tightly contains all elements plus the specified margins.

    Raises:
        ValueError: If elements is empty.
        TypeError: If an element has an unsupported type.

    Example:
        >>> points = [momapy.geometry.Point(0, 0), momapy.geometry.Point(10, 10)]
        >>> bbox = fit(points, xsep=2, ysep=2)
        >>> print(bbox.position)
        Point(x=5.0, y=5.0)
    """
    if not elements:
        raise ValueError("elements must contain at least one element")
    points = []
    for element in elements:
        if momapy.builder.isinstance_or_builder(element, momapy.geometry.Point):
            points.append(element)
        elif momapy.builder.isinstance_or_builder(element, momapy.geometry.Bbox):
            points.append(element.north_west())
            points.append(element.south_east())
        elif momapy.builder.isinstance_or_builder(element, momapy.core.elements.LayoutElement):
            bbox = element.bbox()
            points.append(bbox.north_west())
            points.append(bbox.south_east())
        else:
            raise TypeError(f"{type(element)} not supported")
    bbox = momapy.geometry.Bbox.around_points(points)
    if xsep or ysep:
        bbox = momapy.geometry.Bbox(
            bbox.position,
            bbox.width + 2 * xsep,
            bbox.height + 2 * ysep,
        )
    return bbox


def mid_of(
    obj1: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
) -> momapy.geometry.Point:
    """Return the midpoint between two objects.

    Args:
        obj1: First object (Point, Bbox, Node, or builder).
        obj2: Second object (Point, Bbox, Node, or builder).
            For Bboxes and Nodes, uses their center points.

    Returns:
        A Point at the midpoint between the two objects.

    Example:
        >>> p1 = momapy.geometry.Point(0, 0)
        >>> p2 = momapy.geometry.Point(10, 10)
        >>> mid_of(p1, p2)
        Point(x=5.0, y=5.0)
    """
    if momapy.builder.isinstance_or_builder(
        obj1, (momapy.geometry.Bbox, momapy.core.layout.Node)
    ):
        obj1 = obj1.center()
    if momapy.builder.isinstance_or_builder(
        obj2, (momapy.geometry.Bbox, momapy.core.layout.Node)
    ):
        obj2 = obj2.center()
    segment = momapy.geometry.Segment(obj1, obj2)
    return segment.get_position_at_fraction(0.5)


def cross_vh_of(
    obj1: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
) -> momapy.geometry.Point:
    """Return the point at the vertical cross of obj1 and horizontal cross of obj2.

    Creates a point using obj1's x-coordinate and obj2's y-coordinate.
    Useful for creating elbow connector points.

    Args:
        obj1: Object providing the x-coordinate (Point, Bbox, Node, or builder).
        obj2: Object providing the y-coordinate (Point, Bbox, Node, or builder).
            For Bboxes and Nodes, uses their center points.

    Returns:
        A Point at (obj1.x, obj2.y).

    Example:
        >>> p1 = momapy.geometry.Point(5, 0)
        >>> p2 = momapy.geometry.Point(0, 10)
        >>> cross_vh_of(p1, p2)
        Point(x=5, y=10)
    """
    if momapy.builder.isinstance_or_builder(
        obj1, (momapy.geometry.Bbox, momapy.core.layout.Node)
    ):
        obj1 = obj1.center()
    if momapy.builder.isinstance_or_builder(
        obj2, (momapy.geometry.Bbox, momapy.core.layout.Node)
    ):
        obj2 = obj2.center()
    x = obj1.x
    y = obj2.y
    return momapy.geometry.Point(x, y)


def cross_hv_of(
    obj1: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
) -> momapy.geometry.Point:
    """Return the point at the horizontal cross of obj1 and vertical cross of obj2.

    Creates a point using obj1's y-coordinate and obj2's x-coordinate.
    Useful for creating elbow connector points.

    Args:
        obj1: Object providing the y-coordinate (Point, Bbox, Node, or builder).
        obj2: Object providing the x-coordinate (Point, Bbox, Node, or builder).
            For Bboxes and Nodes, uses their center points.

    Returns:
        A Point at (obj2.x, obj1.y).

    Example:
        >>> p1 = momapy.geometry.Point(0, 5)
        >>> p2 = momapy.geometry.Point(10, 0)
        >>> cross_hv_of(p1, p2)
        Point(x=10, y=5)
    """
    if momapy.builder.isinstance_or_builder(
        obj1, (momapy.geometry.Bbox, momapy.core.layout.Node)
    ):
        obj1 = obj1.center()
    if momapy.builder.isinstance_or_builder(
        obj2, (momapy.geometry.Bbox, momapy.core.layout.Node)
    ):
        obj2 = obj2.center()
    y = obj1.y
    x = obj2.x
    return momapy.geometry.Point(x, y)


def fraction_of(
    arc_layout_element: (momapy.core.layout.SingleHeadedArc | momapy.core.layout.DoubleHeadedArc),
    fraction: float,
) -> tuple[momapy.geometry.Point, float]:
    """Return the position and angle at a given fraction along an arc.

    Args:
        arc_layout_element: An arc layout element (SingleHeadedArc or DoubleHeadedArc).
        fraction: A value between 0.0 and 1.0 representing the position along the arc.

    Returns:
        A tuple containing:
            - The Point at the specified fraction along the arc.
            - The angle (in radians) of the tangent at that position.

    Example:
        >>> arc = some_single_headed_arc
        >>> pos, angle = fraction_of(arc, 0.5)  # Midpoint of arc
    """
    position, angle = arc_layout_element.fraction(fraction)
    return position, angle


def set_position(
    obj: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    position: momapy.geometry.Point | momapy.geometry.PointBuilder,
    anchor: str | None = None,
):
    """Set the position of a builder object.

    Args:
        obj: The builder object to position (NodeBuilder or BboxBuilder).
        position: The target position as a Point or PointBuilder.
        anchor: Optional anchor name to align with the position.
            If specified, the object is shifted so that its anchor point
            aligns with the given position.

    Example:
        >>> builder = momapy.builder.get_or_make_builder_cls(MyNode)()
        >>> set_position(builder, momapy.geometry.Point(100, 100), anchor="center")
    """
    obj.position = position
    if anchor is not None:
        p = getattr(obj, anchor)()
        obj.position += obj.position - p


def set_right_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set obj1's position to the right of obj2 at the specified distance.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance: The horizontal distance from obj2.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = right_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_left_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set obj1's position to the left of obj2 at the specified distance.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance: The horizontal distance from obj2.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = left_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_above_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set obj1's position above obj2 at the specified distance.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance: The vertical distance from obj2.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = above_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_below_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance: float,
    anchor: str | None = None,
):
    """Set obj1's position below obj2 at the specified distance.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance: The vertical distance from obj2.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = below_of(obj2, distance)
    set_position(obj1, position, anchor)


def set_above_left_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set obj1's position above and to the left of obj2.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance1: The vertical distance from obj2.
        distance2: The horizontal distance from obj2. Defaults to distance1 if None.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = above_left_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_above_right_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set obj1's position above and to the right of obj2.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance1: The vertical distance from obj2.
        distance2: The horizontal distance from obj2. Defaults to distance1 if None.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = above_right_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_below_left_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set obj1's position below and to the left of obj2.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance1: The vertical distance from obj2.
        distance2: The horizontal distance from obj2. Defaults to distance1 if None.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = below_left_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_below_right_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ),
    distance1: float,
    distance2: float | None = None,
    anchor: str | None = None,
):
    """Set obj1's position below and to the right of obj2.

    Args:
        obj1: The builder object to position.
        obj2: The reference object to position relative to.
        distance1: The vertical distance from obj2.
        distance2: The horizontal distance from obj2. Defaults to distance1 if None.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = below_right_of(obj2, distance1, distance2)
    set_position(obj1, position, anchor)


def set_fit(
    obj: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    elements: collections.abc.Collection[
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.elements.LayoutElement
        | momapy.core.builders.LayoutElementBuilder
    ],
    xsep: float = 0,
    ysep: float = 0,
    anchor: str | None = None,
):
    """Set obj's dimensions and position to fit the given elements.

    Args:
        obj: The builder object to resize and position.
        elements: Collection of elements to fit within obj.
        xsep: Horizontal margin around the fitted elements.
        ysep: Vertical margin around the fitted elements.
        anchor: Optional anchor name on obj to align with the calculated position.
    """
    bbox = fit(elements, xsep, ysep)
    obj.width = bbox.width
    obj.height = bbox.height
    set_position(obj, bbox.position, anchor)


def set_fraction_of(
    obj: momapy.core.builders.NodeBuilder,
    arc_layout_element: (momapy.core.layout.SingleHeadedArc | momapy.core.layout.DoubleHeadedArc),
    fraction: float,
    anchor: str | None = None,
):
    """Set obj's position and rotation along an arc at the given fraction.

    Args:
        obj: The node builder to position and rotate.
        arc_layout_element: The arc to position along.
        fraction: Position along the arc (0.0 to 1.0).
        anchor: Optional anchor name on obj to align with the calculated position.
    """
    position, angle = fraction_of(arc_layout_element, fraction)
    rotation = momapy.geometry.Rotation(angle, position)
    set_position(obj, position, anchor)
    obj.transform = [rotation]


def set_mid_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    obj3: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    anchor: str | None = None,
):
    """Set obj1's position to the midpoint between obj2 and obj3.

    Args:
        obj1: The builder object to position.
        obj2: First reference object.
        obj3: Second reference object.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = mid_of(obj2, obj3)
    set_position(obj1, position, anchor)


def set_cross_hv_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    obj3: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    anchor: str | None = None,
):
    """Set obj1's position at the horizontal-vertical cross of obj2 and obj3.

    Creates a point using obj2's y-coordinate and obj3's x-coordinate.

    Args:
        obj1: The builder object to position.
        obj2: Object providing the y-coordinate.
        obj3: Object providing the x-coordinate.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = cross_hv_of(obj2, obj3)
    set_position(obj1, position, anchor)


def set_cross_vh_of(
    obj1: momapy.core.builders.NodeBuilder | momapy.geometry.BboxBuilder,
    obj2: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    obj3: (
        momapy.geometry.Point
        | momapy.geometry.PointBuilder
        | momapy.geometry.Bbox
        | momapy.geometry.BboxBuilder
        | momapy.core.layout.Node
        | momapy.core.builders.NodeBuilder
    ),
    anchor: str | None = None,
):
    """Set obj1's position at the vertical-horizontal cross of obj2 and obj3.

    Creates a point using obj2's x-coordinate and obj3's y-coordinate.

    Args:
        obj1: The builder object to position.
        obj2: Object providing the x-coordinate.
        obj3: Object providing the y-coordinate.
        anchor: Optional anchor name on obj1 to align with the calculated position.
    """
    position = cross_vh_of(obj2, obj3)
    set_position(obj1, position, anchor)



"""SBML reader with ReadingContext dataclass."""

import os
import typing
import dataclasses
import collections

import frozendict
import lxml.objectify

import momapy.builder
import momapy.utils
import momapy.io.core
import momapy.sbml.core
import momapy.sbml.io.sbml._parsing
import momapy.sbml.io.sbml._model


@dataclasses.dataclass
class ReadingContext:
    sbml_model: typing.Any
    model: typing.Any
    sbml_id_to_model_element: dict
    sbml_id_to_sbml_element: dict
    map_element_to_annotations: dict
    map_element_to_notes: dict
    map_element_to_ids: dict
    with_annotations: bool
    with_notes: bool


class SBMLReader(momapy.io.core.Reader):

    @staticmethod
    def _register_model_element(
        model_element,
        collection,
        id_,
        id_to_model_element,
    ):
        model_element = momapy.utils.add_or_replace_element_in_set(
            model_element,
            collection,
            func=lambda element, existing_element: element.id_ < existing_element.id_,
        )
        id_to_model_element[id_] = model_element
        return model_element

    @classmethod
    def check_file(cls, file_path: str | os.PathLike):
        try:
            with open(file_path) as f:
                for line in f:
                    if "<sbml " in line:
                        return True
            return False
        except Exception:
            return False

    @classmethod
    def read(
        cls,
        file_path: str | os.PathLike,
        with_annotations=True,
        with_notes=True,
    ) -> momapy.io.core.ReaderResult:
        sbml_document = lxml.objectify.parse(file_path)
        sbml = sbml_document.getroot()
        obj, annotations, notes, ids = cls._make_main_obj(
            sbml_model=sbml.model,
            with_annotations=with_annotations,
            with_notes=with_notes,
        )
        result = momapy.io.core.ReaderResult(
            obj=obj,
            notes=notes,
            annotations=annotations,
            file_path=file_path,
            ids=ids,
        )
        return result

    @classmethod
    def _make_main_obj(
        cls,
        sbml_model,
        with_annotations=True,
        with_notes=True,
    ):
        model = momapy.sbml.io.sbml._model.make_empty_model(sbml_model)
        sbml_id_to_model_element = {}
        map_element_to_annotations = collections.defaultdict(set)
        map_element_to_ids = collections.defaultdict(set)
        map_element_to_notes = collections.defaultdict(set)
        sbml_id_to_sbml_element = momapy.sbml.io.sbml._parsing.make_id_to_element_mapping(
            sbml_model
        )
        ctx = ReadingContext(
            sbml_model=sbml_model,
            model=model,
            sbml_id_to_model_element=sbml_id_to_model_element,
            sbml_id_to_sbml_element=sbml_id_to_sbml_element,
            map_element_to_annotations=map_element_to_annotations,
            map_element_to_notes=map_element_to_notes,
            map_element_to_ids=map_element_to_ids,
            with_annotations=with_annotations,
            with_notes=with_notes,
        )
        for sbml_compartment in momapy.sbml.io.sbml._parsing.get_compartments(ctx.sbml_model):
            cls._make_and_add_compartment(ctx, sbml_compartment)
        for sbml_species in momapy.sbml.io.sbml._parsing.get_species(ctx.sbml_model):
            cls._make_and_add_species(ctx, sbml_species)
        for sbml_reaction in momapy.sbml.io.sbml._parsing.get_reactions(ctx.sbml_model):
            cls._make_and_add_reaction(ctx, sbml_reaction)
        obj = momapy.builder.object_from_builder(model)
        if with_annotations:
            annotations = momapy.sbml.io.sbml._model.make_annotations_from_element(sbml_model)
            map_element_to_annotations[obj].update(annotations)
        if with_notes:
            notes = momapy.sbml.io.sbml._model.make_notes(sbml_model)
            map_element_to_notes[obj].update(notes)
        map_element_to_annotations = frozendict.frozendict(
            {key: frozenset(value) for key, value in map_element_to_annotations.items()}
        )
        map_element_to_notes = frozendict.frozendict(
            {key: frozenset(value) for key, value in map_element_to_notes.items()}
        )
        map_element_to_ids = frozendict.frozendict(
            {key: frozenset(value) for key, value in map_element_to_ids.items()}
        )
        return (
            obj,
            map_element_to_annotations,
            map_element_to_notes,
            map_element_to_ids,
        )

    @classmethod
    def _make_and_add_compartment(cls, ctx, sbml_compartment):
        model_element = momapy.sbml.io.sbml._model.make_compartment(sbml_compartment, ctx.model)
        model_element = momapy.builder.object_from_builder(model_element)
        model_element = cls._register_model_element(
            model_element,
            ctx.model.compartments,
            sbml_compartment.get("id"),
            ctx.sbml_id_to_model_element,
        )
        ctx.map_element_to_ids[model_element].add(sbml_compartment.get("id"))
        if ctx.with_annotations:
            annotations = momapy.sbml.io.sbml._model.make_annotations_from_element(sbml_compartment)
            if annotations:
                ctx.map_element_to_annotations[model_element].update(annotations)
        if ctx.with_notes:
            notes = momapy.sbml.io.sbml._model.make_notes(sbml_compartment)
            ctx.map_element_to_notes[model_element].update(notes)
        return model_element

    @classmethod
    def _make_and_add_species(cls, ctx, sbml_species):
        model_element = momapy.sbml.io.sbml._model.make_species(sbml_species, ctx.model, ctx.sbml_id_to_model_element)
        model_element = momapy.builder.object_from_builder(model_element)
        model_element = cls._register_model_element(
            model_element,
            ctx.model.species,
            sbml_species.get("id"),
            ctx.sbml_id_to_model_element,
        )
        if ctx.with_annotations:
            annotations = momapy.sbml.io.sbml._model.make_annotations_from_element(sbml_species)
            if annotations:
                ctx.map_element_to_annotations[model_element].update(annotations)
        if ctx.with_notes:
            notes = momapy.sbml.io.sbml._model.make_notes(sbml_species)
            ctx.map_element_to_notes[model_element].update(notes)
        return model_element

    @classmethod
    def _make_and_add_reaction(cls, ctx, sbml_reaction):
        model_element = momapy.sbml.io.sbml._model.make_reaction(sbml_reaction, ctx.model)
        for sbml_reactant in momapy.sbml.io.sbml._parsing.get_reactants(sbml_reaction):
            cls._make_and_add_reactant(ctx, sbml_reactant, model_element)
        for sbml_product in momapy.sbml.io.sbml._parsing.get_products(sbml_reaction):
            cls._make_and_add_product(ctx, sbml_product, model_element)
        for sbml_modifier in momapy.sbml.io.sbml._parsing.get_modifiers(sbml_reaction):
            cls._make_and_add_modifier(ctx, sbml_modifier, model_element)
        model_element = momapy.builder.object_from_builder(model_element)
        model_element = cls._register_model_element(
            model_element,
            ctx.model.reactions,
            sbml_reaction.get("id"),
            ctx.sbml_id_to_model_element,
        )
        ctx.map_element_to_ids[model_element].add(sbml_reaction.get("id"))
        if ctx.with_annotations:
            annotations = momapy.sbml.io.sbml._model.make_annotations_from_element(sbml_reaction)
            if annotations:
                ctx.map_element_to_annotations[model_element].update(annotations)
            if ctx.with_notes:
                notes = momapy.sbml.io.sbml._model.make_notes(sbml_reaction)
                ctx.map_element_to_notes[model_element].update(notes)
        return model_element

    @classmethod
    def _make_and_add_reactant(cls, ctx, sbml_species_reference, super_model_element):
        model_element = momapy.sbml.io.sbml._model.make_species_reference(sbml_species_reference, ctx.model, ctx.sbml_id_to_model_element)
        super_model_element.reactants.add(model_element)
        ctx.sbml_id_to_model_element[model_element.id_] = model_element
        return model_element

    @classmethod
    def _make_and_add_product(cls, ctx, sbml_species_reference, super_model_element):
        model_element = momapy.sbml.io.sbml._model.make_species_reference(sbml_species_reference, ctx.model, ctx.sbml_id_to_model_element)
        super_model_element.products.add(model_element)
        ctx.sbml_id_to_model_element[model_element.id_] = model_element
        return model_element

    @classmethod
    def _make_and_add_modifier(cls, ctx, sbml_modifier_species_reference, super_model_element):
        model_element = momapy.sbml.io.sbml._model.make_modifier_species_reference(sbml_modifier_species_reference, ctx.model, ctx.sbml_id_to_model_element)
        super_model_element.modifiers.add(model_element)
        ctx.sbml_id_to_model_element[model_element.id_] = model_element
        return model_element



"""Tests for momapy.builder module."""
import pytest
import dataclasses
import momapy.builder


@dataclasses.dataclass(frozen=True)
class SimpleClass:
    """Simple dataclass for testing builders."""
    value: int
    name: str


def test_isinstance_or_builder():
    """Test isinstance_or_builder function."""
    obj = SimpleClass(42, "test")
    assert momapy.builder.isinstance_or_builder(obj, SimpleClass)
    assert not momapy.builder.isinstance_or_builder("string", SimpleClass)


def test_get_or_make_builder_cls():
    """Test get_or_make_builder_cls function."""
    # Test with a simple dataclass
    builder_cls = momapy.builder.get_or_make_builder_cls(SimpleClass)

    # Builder class should be created
    assert builder_cls is not None
    assert issubclass(builder_cls, momapy.builder.Builder)


def test_builder_from_object():
    """Test creating a builder from an object."""
    obj = SimpleClass(42, "test")
    builder_cls = momapy.builder.get_or_make_builder_cls(SimpleClass)
    builder = builder_cls.from_object(obj)

    assert isinstance(builder, momapy.builder.Builder)
    assert builder.value == 42
    assert builder.name == "test"


def test_builder_build():
    """Test building an object from a builder."""
    obj = SimpleClass(42, "test")
    builder_cls = momapy.builder.get_or_make_builder_cls(SimpleClass)
    builder = builder_cls.from_object(obj)

    # Build back to object
    rebuilt_obj = builder.build()

    assert isinstance(rebuilt_obj, SimpleClass)
    assert rebuilt_obj.value == 42
    assert rebuilt_obj.name == "test"


def test_builder_from_object_function():
    """Test builder_from_object utility function."""
    obj = SimpleClass(42, "test")
    builder = momapy.builder.builder_from_object(obj)

    assert isinstance(builder, momapy.builder.Builder)


def test_object_from_builder_function():
    """Test object_from_builder utility function."""
    obj = SimpleClass(42, "test")
    builder = momapy.builder.builder_from_object(obj)
    rebuilt_obj = momapy.builder.object_from_builder(builder)

    assert isinstance(rebuilt_obj, SimpleClass)
    assert rebuilt_obj.value == 42
    assert rebuilt_obj.name == "test"


def test_builder_registry():
    """Test that builders dictionary exists."""
    assert isinstance(momapy.builder.builders, dict)



"""Standalone model-building functions for SBGN-ML reader."""

import lxml.etree

import momapy.builder
import momapy.core.elements
import momapy.sbgn.pd
import momapy.sbml.core
import momapy.sbgn.io.sbgnml._parsing
import momapy.sbml.io.sbml._parsing
import momapy.sbml.io.sbml._qualifiers


def make_annotations(sbgnml_element):
    annotations = []
    sbgnml_rdf = momapy.sbgn.io.sbgnml._parsing.get_rdf(sbgnml_element)
    if sbgnml_rdf is not None:
        description = momapy.sbml.io.sbml._parsing.get_description(sbgnml_rdf)
        if description is not None:
            for bq_element in description.getchildren():
                key = momapy.sbml.io.sbml._parsing.get_prefix_and_name(bq_element.tag)
                qualifier = momapy.sbml.io.sbml._qualifiers.QUALIFIER_ATTRIBUTE_TO_QUALIFIER_MEMBER.get(key)
                if qualifier is not None:
                    bags = momapy.sbml.io.sbml._parsing.get_bags(bq_element)
                    for bag in bags:
                        lis = momapy.sbml.io.sbml._parsing.get_list_items(bag)
                        resources = [
                            li.get(
                                f"{{{momapy.sbml.io.sbml._parsing._RDF_NAMESPACE}}}resource"
                            )
                            for li in lis
                        ]
                        annotation = momapy.sbml.core.RDFAnnotation(
                            qualifier=qualifier,
                            resources=frozenset(resources),
                        )
                        annotations.append(annotation)
    return annotations


def make_notes(sbgnml_element):
    sbgnml_notes = momapy.sbgn.io.sbgnml._parsing.get_notes(sbgnml_element)
    if sbgnml_notes is not None:
        for child_element in sbgnml_notes.iterchildren():
            break
        return [lxml.etree.tostring(child_element, encoding="unicode")]
    return []


def set_label(model_element, sbgnml_element):
    sbgnml_label = getattr(sbgnml_element, "label", None)
    if sbgnml_label is not None:
        model_element.label = sbgnml_label.get("text")


def set_compartment(
    model_element, sbgnml_element, sbgnml_id_to_model_element
):
    sbgnml_compartment_ref = sbgnml_element.get("compartmentRef")
    if sbgnml_compartment_ref is not None:
        model_element.compartment = sbgnml_id_to_model_element[sbgnml_compartment_ref]


def set_stoichiometry(model_element, sbgnml_stoichiometry):
    if sbgnml_stoichiometry is None:
        return
    sbgnml_label = getattr(sbgnml_stoichiometry, "label", None)
    if sbgnml_label is not None:
        sbgnml_stoichiometry_text = sbgnml_label.get("text")
        try:
            stoichiometry = float(sbgnml_stoichiometry_text)
        except ValueError:
            stoichiometry = sbgnml_stoichiometry_text
        model_element.stoichiometry = float(stoichiometry)


def make_compartment(sbgnml_compartment, model, module):
    model_element = model.new_element(module.Compartment)
    model_element.id_ = sbgnml_compartment.get("id")
    set_label(model_element, sbgnml_compartment)
    return model_element


def make_entity_pool_or_subunit(
    sbgnml_entity_pool_or_subunit, model, model_element_cls, sbgnml_id_to_model_element
):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_entity_pool_or_subunit.get("id")
    set_compartment(
        model_element, sbgnml_entity_pool_or_subunit, sbgnml_id_to_model_element
    )
    set_label(model_element, sbgnml_entity_pool_or_subunit)
    return model_element


def make_activity(
    sbgnml_activity, model, model_element_cls, sbgnml_id_to_model_element
):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_activity.get("id")
    set_compartment(
        model_element, sbgnml_activity, sbgnml_id_to_model_element
    )
    set_label(model_element, sbgnml_activity)
    return model_element


def make_state_variable(sbgnml_state_variable, model, order=None):
    sbgnml_id = sbgnml_state_variable.get("id")
    sbgnml_state = getattr(sbgnml_state_variable, "state", None)
    if sbgnml_state is None:
        value = None
        variable = None
    else:
        sbgnml_value = sbgnml_state.get("value")
        if sbgnml_value:
            value = sbgnml_value
        else:
            value = None
        sbgnml_variable = sbgnml_state.get("variable")
        variable = sbgnml_variable
    model_element = model.new_element(momapy.sbgn.pd.StateVariable)
    model_element.id_ = sbgnml_id
    model_element.value = value
    model_element.variable = variable
    model_element.order = order
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element


def make_unit_of_information(sbgnml_unit_of_information, model, model_element_cls):
    sbgnml_label = getattr(sbgnml_unit_of_information, "label", None)
    sbgnml_id = sbgnml_unit_of_information.get("id")
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_id
    if sbgnml_label is not None:
        split_label = sbgnml_label.get("text").split(":")
        model_element.value = split_label[-1]
        if len(split_label) > 1:
            model_element.prefix = split_label[0]
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element


def make_submap(sbgnml_submap, model, model_element_cls):
    sbgnml_id = sbgnml_submap.get("id")
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_id
    set_label(model_element, sbgnml_submap)
    return model_element


def make_terminal_or_tag(sbgnml_terminal_or_tag, model, is_terminal):
    sbgnml_id = sbgnml_terminal_or_tag.get("id")
    if is_terminal:
        model_element_cls = momapy.sbgn.pd.Terminal
    else:
        model_element_cls = momapy.sbgn.pd.Tag
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_id
    set_label(model_element, sbgnml_terminal_or_tag)
    return model_element


def make_reference(
    sbgnml_equivalence_arc, model, sbgnml_id_to_model_element, is_terminal
):
    sbgnml_id = sbgnml_equivalence_arc.get("id")
    # For terminals and tags, equivalence arc goes from the referred node
    # to the terminal or tag. We invert the arc, so that the arc goes
    # from the reference to the referred node.
    sbgnml_target_id = sbgnml_equivalence_arc.get("source")
    if is_terminal:
        model_element_cls = momapy.sbgn.pd.TerminalReference
    else:
        model_element_cls = momapy.sbgn.pd.TagReference
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_id
    target_model_element = sbgnml_id_to_model_element[sbgnml_target_id]
    model_element.element = target_model_element
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element


def make_stoichiometric_process(
    sbgnml_process, model, model_element_cls, sbgnml_glyph_id_to_sbgnml_arcs
):
    sbgnml_id = sbgnml_process.get("id")
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_id
    model_element.reversible = momapy.sbgn.io.sbgnml._parsing.is_process_reversible(
        sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs
    )
    return model_element


def make_reactant(sbgnml_consumption_arc, model, sbgnml_id_to_model_element):
    sbgnml_source_id = sbgnml_consumption_arc.get("source")
    sbgnml_stoichiometry = momapy.sbgn.io.sbgnml._parsing.get_stoichiometry(
        sbgnml_consumption_arc
    )
    model_element = model.new_element(momapy.sbgn.pd.Reactant)
    model_element.id_ = sbgnml_consumption_arc.get("id")
    source_model_element = sbgnml_id_to_model_element[sbgnml_source_id]
    model_element.element = source_model_element
    set_stoichiometry(model_element, sbgnml_stoichiometry)
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element


def make_product(
    sbgnml_production_arc,
    model,
    sbgnml_id_to_model_element,
    super_model_element,
    super_sbgnml_element,
    process_direction,
):
    sbgnml_target_id = sbgnml_production_arc.get("target")
    sbgnml_stoichiometry = momapy.sbgn.io.sbgnml._parsing.get_stoichiometry(
        sbgnml_production_arc
    )
    if super_model_element.reversible:
        if process_direction == momapy.core.elements.Direction.HORIZONTAL:
            if float(sbgnml_production_arc.start.get("x")) > float(
                super_sbgnml_element.bbox.get("x")
            ):  # RIGHT
                model_element_cls = momapy.sbgn.pd.Product
            else:
                model_element_cls = momapy.sbgn.pd.Reactant  # LEFT
        else:
            if float(sbgnml_production_arc.start.get("y")) > float(
                super_sbgnml_element.bbox.get("y")
            ):  # TOP
                model_element_cls = momapy.sbgn.pd.Product
            else:
                model_element_cls = momapy.sbgn.pd.Reactant  # BOTTOM
    else:
        model_element_cls = momapy.sbgn.pd.Product
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_production_arc.get("id")
    target_model_element = sbgnml_id_to_model_element[sbgnml_target_id]
    model_element.element = target_model_element
    set_stoichiometry(model_element, sbgnml_stoichiometry)
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element


def make_logical_operator(sbgnml_logical_operator, model, model_element_cls):
    sbgnml_id = sbgnml_logical_operator.get("id")
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_id
    return model_element


def make_logical_operator_input(sbgnml_logic_arc, model, source_model_element):
    model_element = model.new_element(momapy.sbgn.pd.LogicalOperatorInput)
    model_element.id_ = sbgnml_logic_arc.get("id")
    model_element.element = source_model_element
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element


def make_modulation(
    sbgnml_modulation, model, model_element_cls, source_model_element, target_model_element
):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = sbgnml_modulation.get("id")
    model_element.source = source_model_element
    model_element.target = target_model_element
    model_element = momapy.builder.object_from_builder(model_element)
    return model_element



"""SBGN-ML XML parsing utilities.

Pure XML traversal and data-extraction functions for SBGN-ML documents.
No momapy object construction happens here: all functions work with lxml
objectify trees and return raw Python / geometry values.
"""

import momapy.core.elements
import momapy.sbml.io.sbml._parsing


def transform_class(sbgnml_class):
    return sbgnml_class.upper().replace(" ", "_")


def has_undefined_variable(sbgnml_state_variable):
    sbgnml_state = getattr(sbgnml_state_variable, "state", None)
    if sbgnml_state is None:
        return True
    sbgnml_variable = sbgnml_state.get("variable")
    return sbgnml_variable is None


def get_glyphs(sbgnml_element):
    return list(getattr(sbgnml_element, "glyph", []))


def get_glyphs_recursively(sbgnml_element):
    sub_glyphs = []
    for sub_glyph in get_glyphs(sbgnml_element):
        sub_glyphs.append(sub_glyph)
        sub_glyphs += get_glyphs_recursively(sub_glyph)
    return sub_glyphs


def get_arcs(sbgnml_element):
    return list(getattr(sbgnml_element, "arc", []))


def get_ports(sbgnml_element):
    return list(getattr(sbgnml_element, "port", []))


def get_nexts(sbgnml_arc):
    return list(getattr(sbgnml_arc, "next", []))


def get_sbgnml_points(sbgnml_arc):
    return (
        [sbgnml_arc.start]
        + get_nexts(sbgnml_arc)
        + [sbgnml_arc.end]
    )



def get_annotation(sbgnml_element):
    extension = getattr(sbgnml_element, "extension", None)
    if extension is None:
        return None
    annotation = getattr(extension, "annotation", None)
    if annotation is None:
        return getattr(
            extension, "{}annotation", None
        )  # to account for bug in libsbgn: no namespace
    return annotation


def get_notes(sbgnml_element):
    return getattr(sbgnml_element, "notes", None)




def get_rdf(sbgnml_element):
    annotation = get_annotation(sbgnml_element)
    if annotation is None:
        return None
    return getattr(annotation, f"{{{momapy.sbml.io.sbml._parsing._RDF_NAMESPACE}}}RDF", None)


_SBGNML_STATE_VARIABLE_CLASSES = {"state variable"}

_SBGNML_UNIT_OF_INFORMATION_CLASSES = {"unit of information"}

_SBGNML_TERMINAL_CLASSES = {"terminal"}

_SBGNML_SUBUNIT_CLASSES = {
    "unspecified entity",
    "macromolecule",
    "macromolecule multimer",
    "simple chemical",
    "simple chemical multimer",
    "nucleic acid feature",
    "nucleic acid feature multimer",
    "complex",
    "complex multimer",
}


def get_state_variables(sbgnml_element):
    return [
        subglyph
        for subglyph in get_glyphs(sbgnml_element)
        if subglyph.get("class") in _SBGNML_STATE_VARIABLE_CLASSES
    ]


def get_units_of_information(sbgnml_element):
    return [
        subglyph
        for subglyph in get_glyphs(sbgnml_element)
        if subglyph.get("class") in _SBGNML_UNIT_OF_INFORMATION_CLASSES
    ]


def get_subunits(sbgnml_element):
    return [
        subglyph
        for subglyph in get_glyphs(sbgnml_element)
        if subglyph.get("class") in _SBGNML_SUBUNIT_CLASSES
    ]


def get_terminals(sbgnml_element):
    return [
        subglyph
        for subglyph in get_glyphs(sbgnml_element)
        if subglyph.get("class") in _SBGNML_TERMINAL_CLASSES
    ]


def get_stoichiometry(sbgnml_element):
    for sbgnml_subglyph in get_glyphs(sbgnml_element):
        if sbgnml_subglyph.get("class") == "stoichiometry":
            return sbgnml_subglyph
    return None


def get_consumption_and_production_arcs(
    sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs
):
    sbgnml_consumption_arcs = []
    sbgnml_production_arcs = []
    for sbgnml_arc in sbgnml_glyph_id_to_sbgnml_arcs[sbgnml_process.get("id")]:
        if transform_class(sbgnml_arc.get("class")) == "CONSUMPTION":
            sbgnml_consumption_arcs.append(sbgnml_arc)
        elif transform_class(sbgnml_arc.get("class")) == "PRODUCTION":
            sbgnml_production_arcs.append(sbgnml_arc)
    return sbgnml_consumption_arcs, sbgnml_production_arcs


def get_equivalence_arcs(
    sbgnml_tag_or_terminal,
    sbgnml_id_to_sbgnml_element,
    sbgnml_glyph_id_to_sbgnml_arcs,
):
    sbgnml_equivalence_arcs = []
    for sbgnml_arc in sbgnml_glyph_id_to_sbgnml_arcs[
        sbgnml_tag_or_terminal.get("id")
    ]:
        if (
            transform_class(sbgnml_arc.get("class")) == "EQUIVALENCE_ARC"
            and sbgnml_id_to_sbgnml_element[sbgnml_arc.get("target")]
            == sbgnml_tag_or_terminal
        ):
            sbgnml_equivalence_arcs.append(sbgnml_arc)
    return sbgnml_equivalence_arcs


def get_logic_arcs(
    sbgnml_operator,
    sbgnml_id_to_sbgnml_element,
    sbgnml_glyph_id_to_sbgnml_arcs,
):
    sbgnml_logic_arcs = []
    for sbgnml_arc in sbgnml_glyph_id_to_sbgnml_arcs[sbgnml_operator.get("id")]:
        if (
            transform_class(sbgnml_arc.get("class")) == "LOGIC_ARC"
            and sbgnml_id_to_sbgnml_element[sbgnml_arc.get("target")]
            == sbgnml_operator
        ):
            sbgnml_logic_arcs.append(sbgnml_arc)
    return sbgnml_logic_arcs


def get_process_direction(sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs):
    for sbgnml_port in get_ports(sbgnml_process):
        if float(sbgnml_port.get("x")) < float(
            sbgnml_process.bbox.get("x")
        ) or float(sbgnml_port.get("x")) >= float(
            sbgnml_process.bbox.get("x")
        ) + float(
            sbgnml_process.bbox.get("w")
        ):  # LEFT OR RIGHT
            return momapy.core.elements.Direction.HORIZONTAL
        else:
            return momapy.core.elements.Direction.VERTICAL
    return momapy.core.elements.Direction.VERTICAL  # default is vertical


def get_direction(sbgnml_element):
    sbgnml_orientation = sbgnml_element.get("orientation")
    if sbgnml_orientation is None:
        return momapy.core.elements.Direction.RIGHT
    orientation = transform_class(sbgnml_orientation)
    return momapy.core.elements.Direction[orientation]


def is_operator_left_to_right(
    sbgnml_operator,
    sbgnml_id_to_sbgnml_element,
    sbgnml_glyph_id_to_sbgnml_arcs,
):
    sbgnml_logic_arcs = get_logic_arcs(
        sbgnml_operator,
        sbgnml_id_to_sbgnml_element,
        sbgnml_glyph_id_to_sbgnml_arcs,
    )
    operator_direction = get_process_direction(
        sbgnml_operator, sbgnml_glyph_id_to_sbgnml_arcs
    )
    for sbgnml_logic_arc in sbgnml_logic_arcs:
        if operator_direction == momapy.core.elements.Direction.HORIZONTAL:
            if float(sbgnml_logic_arc.end.get("x")) < float(
                sbgnml_operator.bbox.get("x")
            ):
                return True
            else:
                return False
        else:
            if float(sbgnml_logic_arc.end.get("y")) < float(
                sbgnml_operator.bbox.get("y")
            ):
                return True
            else:
                return False
    return True


def is_process_left_to_right(sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs):
    process_direction = get_process_direction(
        sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs
    )
    sbgnml_consumption_arcs, sbgnml_production_arcs = (
        get_consumption_and_production_arcs(
            sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs
        )
    )
    if sbgnml_production_arcs:
        if not sbgnml_production_arcs:  # process is reversible
            return True  # defaults to left to right
        sbgnml_production_arc = sbgnml_production_arcs[0]
        if process_direction == momapy.core.elements.Direction.HORIZONTAL:
            if float(sbgnml_production_arc.start.get("x")) >= float(
                sbgnml_process.bbox.get("x")
            ):
                return True
            else:
                return False
        else:
            if float(sbgnml_production_arc.start.get("y")) >= float(
                sbgnml_process.bbox.get("y")
            ):
                return True
            return False
    if sbgnml_consumption_arcs:
        sbgnml_consumption_arc = sbgnml_consumption_arcs[0]
        if process_direction == momapy.core.elements.Direction.HORIZONTAL:
            if float(sbgnml_consumption_arc.end.get("x")) <= float(
                sbgnml_process.bbox.get("x")
            ):
                return True
            else:
                return False
        else:
            if float(sbgnml_consumption_arc.end.get("y")) <= float(
                sbgnml_process.bbox.get("y")
            ):
                return True
            return False


def is_process_reversible(sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs):
    sbgnml_consumption_arcs, _ = (
        get_consumption_and_production_arcs(
            sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs
        )
    )
    if sbgnml_consumption_arcs:
        return False
    return True


def get_connectors_length(sbgnml_process):
    left_connector_length = None
    right_connector_length = None
    sbgnml_bbox = sbgnml_process.bbox
    sbgnml_bbox_x = float(sbgnml_bbox.get("x"))
    sbgnml_bbox_y = float(sbgnml_bbox.get("y"))
    sbgnml_bbox_w = float(sbgnml_bbox.get("w"))
    sbgnml_bbox_h = float(sbgnml_bbox.get("h"))
    for sbgnml_port in get_ports(sbgnml_process):
        sbgnml_port_x = float(sbgnml_port.get("x"))
        sbgnml_port_y = float(sbgnml_port.get("y"))
        if sbgnml_port_x < sbgnml_bbox_x:  # LEFT
            left_connector_length = sbgnml_bbox_x - sbgnml_port_x
        elif sbgnml_port_y < sbgnml_bbox_y:  # UP
            left_connector_length = sbgnml_bbox_y - sbgnml_port_y
        elif sbgnml_port_x >= sbgnml_bbox_x + sbgnml_bbox_w:  # RIGHT
            right_connector_length = sbgnml_port_x - sbgnml_bbox_x - sbgnml_bbox_w
        elif sbgnml_port_y >= sbgnml_bbox_y + sbgnml_bbox_h:  # DOWN
            right_connector_length = sbgnml_port_y - sbgnml_bbox_y - sbgnml_bbox_h
    return left_connector_length, right_connector_length
